#!/usr/bin/env python3

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.zephyr.errno import errno


class data_logger_erase(InfuseRpcCommand, defs.data_logger_erase):
    @classmethod
    def add_parser(cls, parser):
        logger = parser.add_mutually_exclusive_group(required=True)
        logger.add_argument("--onboard", action="store_true", help="Onboard flash logger")
        logger.add_argument("--removable", action="store_true", help="Removable flash logger (SD)")
        parser.add_argument(
            "--erase-all",
            action="store_true",
            help="Erase entire address space, not just written blocks",
        )

    def __init__(self, args):
        self.infuse_id = args.id
        if args.onboard:
            self.logger = defs.rpc_enum_data_logger.FLASH_ONBOARD
        elif args.removable:
            self.logger = defs.rpc_enum_data_logger.FLASH_REMOVABLE
        else:
            raise NotImplementedError
        self.erase_all = 1 if args.erase_all else 0

    def request_struct(self):
        return self.request(self.logger, self.erase_all)

    def request_json(self):
        return {"logger": self.logger.name, "erase_empty": self.erase_all}

    def handle_response(self, return_code, _response):
        if return_code != 0:
            print(f"Failed to erase data logger ({errno.strerror(-return_code)})")
            return

        print("Data logger erased")



#!/usr/bin/env python3

import ctypes
import random
from collections.abc import Callable

from infuse_iot import rpc
from infuse_iot.common import InfuseType
from infuse_iot.epacket.packet import Auth, PacketOutput, PacketReceived
from infuse_iot.socket_comms import (
    ClientNotification,
    ClientNotificationConnectionDropped,
    ClientNotificationEpacketReceived,
    GatewayRequestEpacketSend,
    LocalClient,
)


class RpcClient:
    def __init__(
        self,
        client: LocalClient,
        max_payload: int,
        infuse_id: int,
        rx_cb: Callable[[ClientNotification], None] | None = None,
    ):
        self._request_id = random.randint(0, 2**31 - 1)
        self._client = client
        self._id = infuse_id
        self._max_payload = max_payload
        self._rx_cb = rx_cb

    def _finalise_command(
        self, rpc_rsp: PacketReceived, rsp_decoder: Callable[[bytes], ctypes.LittleEndianStructure]
    ) -> tuple[rpc.ResponseHeader, ctypes.LittleEndianStructure | None]:
        # Convert response bytes back to struct form
        rsp_header = rpc.ResponseHeader.from_buffer_copy(rpc_rsp.payload)
        rsp_payload = rpc_rsp.payload[ctypes.sizeof(rpc.ResponseHeader) :]
        try:
            rsp_data = rsp_decoder(rsp_payload)
        except ValueError:
            rsp_data = None
        return (rsp_header, rsp_data)

    def _client_recv(self) -> ClientNotification | None:
        rsp = self._client.receive()
        if rsp is not None and self._rx_cb is not None:
            self._rx_cb(rsp)
        return rsp

    def _wait_data_ack(self) -> PacketReceived:
        while True:
            rsp = self._client_recv()
            if rsp is None:
                continue
            if not isinstance(rsp, ClientNotificationEpacketReceived):
                continue
            if rsp.epacket.ptype == InfuseType.RPC_RSP:
                rsp_header = rpc.ResponseHeader.from_buffer_copy(rsp.epacket.payload)
                if rsp_header.request_id == self._request_id:
                    return rsp.epacket
            elif rsp.epacket.ptype != InfuseType.RPC_DATA_ACK:
                continue
            data_ack = rpc.DataAck.from_buffer_copy(rsp.epacket.payload)
            # Response to the request we sent
            if data_ack.request_id != self._request_id:
                continue
            return rsp.epacket

    def _wait_rpc_rsp(self) -> PacketReceived:
        # Wait for responses
        while True:
            rsp = self._client_recv()
            if rsp is None:
                continue
            if not isinstance(rsp, ClientNotificationEpacketReceived):
                continue
            # RPC response packet
            if rsp.epacket.ptype != InfuseType.RPC_RSP:
                continue
            rsp_header = rpc.ResponseHeader.from_buffer_copy(rsp.epacket.payload)
            # Response to the request we sent
            if rsp_header.request_id != self._request_id:
                continue
            return rsp.epacket

    def _run_data_send_core(
        self,
        cmd_id: int,
        auth: Auth,
        params: bytes,
        data: list[bytes],
        total_size: int,
        packet_idx: bool,
        progress_cb: Callable[[int], None] | None,
        rsp_decoder: Callable[[bytes], ctypes.LittleEndianStructure],
    ) -> tuple[rpc.ResponseHeader, ctypes.LittleEndianStructure | None]:
        self._request_id += 1
        ack_period = 1
        header = rpc.RequestHeader(self._request_id, cmd_id)  # type: ignore
        data_hdr = rpc.RequestDataHeader(total_size, ack_period)

        request_packet = bytes(header) + bytes(data_hdr) + params
        pkt = PacketOutput(
            self._id,
            auth,
            InfuseType.RPC_CMD,
            request_packet,
        )
        req = GatewayRequestEpacketSend(pkt)
        self._client.send(req)

        # Wait for initial ACK
        recv = self._wait_data_ack()
        if recv.ptype == InfuseType.RPC_RSP:
            return self._finalise_command(recv, rsp_decoder)

        # Send data payloads chunked as requested
        ack_cnt = -ack_period
        offset = 0
        for chunk_id, chunk in enumerate(data):
            hdr = rpc.DataHeader(self._request_id, chunk_id if packet_idx else offset)
            pkt_bytes = bytes(hdr) + chunk
            pkt = PacketOutput(
                self._id,
                auth,
                InfuseType.RPC_DATA,
                pkt_bytes,
            )
            self._client.send(GatewayRequestEpacketSend(pkt))
            ack_cnt += 1

            # Wait for ACKs at the period
            if ack_cnt == ack_period:
                recv = self._wait_data_ack()
                if recv.ptype == InfuseType.RPC_RSP:
                    return self._finalise_command(recv, rsp_decoder)
                ack_cnt = 0

            offset += len(chunk)
            if progress_cb:
                progress_cb(chunk_id + 1 if packet_idx else offset)

        recv = self._wait_rpc_rsp()
        return self._finalise_command(recv, rsp_decoder)

    def run_data_send_cmd(
        self,
        cmd_id: int,
        auth: Auth,
        params: bytes,
        data: bytes,
        progress_cb: Callable[[int], None] | None,
        rsp_decoder: Callable[[bytes], ctypes.LittleEndianStructure],
    ) -> tuple[rpc.ResponseHeader, ctypes.LittleEndianStructure | None]:
        # Maxmimum payload size of interface
        size = self._max_payload - ctypes.sizeof(rpc.DataHeader)
        # Round payload down to multiple of 4 bytes
        size -= size % 4
        # itertools.batched once Python 3.12 is the minimum version
        chunks = [data[i : i + size] for i in range(0, len(data), size)]
        # Run with pre-computed chunks
        return self._run_data_send_core(cmd_id, auth, params, chunks, len(data), False, progress_cb, rsp_decoder)

    def run_data_send_cmd_chunked(
        self,
        cmd_id: int,
        auth: Auth,
        params: bytes,
        data: list[bytes],
        progress_cb: Callable[[int], None] | None,
        rsp_decoder: Callable[[bytes], ctypes.LittleEndianStructure],
    ) -> tuple[rpc.ResponseHeader, ctypes.LittleEndianStructure | None]:
        return self._run_data_send_core(cmd_id, auth, params, data, len(data), True, progress_cb, rsp_decoder)

    def run_data_recv_cmd(
        self,
        cmd_id: int,
        auth: Auth,
        params: bytes,
        size: int,
        recv_cb: Callable[[int, bytes], None],
        rsp_decoder: Callable[[bytes], ctypes.LittleEndianStructure],
    ) -> tuple[rpc.ResponseHeader, ctypes.LittleEndianStructure | None]:
        self._request_id += 1
        header = rpc.RequestHeader(self._request_id, cmd_id)
        data_hdr = rpc.RequestDataHeader(size, 0)

        request_packet = bytes(header) + bytes(data_hdr) + params
        pkt = PacketOutput(
            self._id,
            auth,
            InfuseType.RPC_CMD,
            request_packet,
        )
        req = GatewayRequestEpacketSend(pkt)
        self._client.send(req)

        while True:
            rsp = self._client_recv()
            if rsp is None:
                continue
            if isinstance(rsp, ClientNotificationConnectionDropped):
                raise ConnectionAbortedError
            if not isinstance(rsp, ClientNotificationEpacketReceived):
                continue
            if rsp.epacket.ptype == InfuseType.RPC_RSP:
                rsp_header = rpc.ResponseHeader.from_buffer_copy(rsp.epacket.payload)
                # Response to the request we sent
                if rsp_header.request_id != self._request_id:
                    continue
                # Convert response bytes back to struct form
                rsp_payload = rsp.epacket.payload[ctypes.sizeof(rpc.ResponseHeader) :]
                rsp_data = rsp_decoder(rsp_payload)
                return (rsp_header, rsp_data)

            if rsp.epacket.ptype != InfuseType.RPC_DATA:
                continue
            data = rpc.DataHeader.from_buffer_copy(rsp.epacket.payload)
            # Response to the request we sent
            if data.request_id != self._request_id:
                continue

            recv_cb(data.offset, rsp.epacket.payload[ctypes.sizeof(rpc.DataHeader) :])

    def run_standard_cmd(
        self, cmd_id: int, auth: Auth, params: bytes, rsp_decoder: Callable[[bytes], ctypes.LittleEndianStructure]
    ) -> tuple[rpc.ResponseHeader, ctypes.LittleEndianStructure | None]:
        self._request_id += 1
        header = rpc.RequestHeader(self._request_id, cmd_id)  # type: ignore

        request_packet = bytes(header) + params
        pkt = PacketOutput(
            self._id,
            auth,
            InfuseType.RPC_CMD,
            request_packet,
        )
        req = GatewayRequestEpacketSend(pkt)
        self._client.send(req)
        recv = self._wait_rpc_rsp()
        return self._finalise_command(recv, rsp_decoder)



#!/usr/bin/env python3

import enum


class OperationalState(enum.IntEnum):
    UNKNOWN = 0
    NOTPRESENT = 1
    DOWN = 2
    LOWERLAYERDOWN = 3
    TESTING = 4
    DORMANT = 5
    UP = 6


class InterfaceFlags(enum.Flag):
    UP = 0x0001
    POINTOPOINT = 0x0002
    PROMISC = 0x0004
    NO_AUTO_START = 0x0008
    SUSPENDED = 0x0010
    FORWARD_MULTICASTS = 0x0020
    IPV4 = 0x0040
    IPV6 = 0x0080
    RUNNING = 0x0100
    LOWER_UP = 0x0200
    DORMANT = 0x0400
    IPV6_NO_ND = 0x0800
    IPV6_NO_MLD = 0x1000
    NO_TX_LOCK = 0x2000


class L2Flags(enum.Flag):
    MULTICAST = 0x0001
    MULTICAST_SKIP_SOLICIT = 0x0002
    PROMISC_MODE = 0x0004
    POINT_TO_POINT = 0x0008



#!/usr/bin/env python3

import ctypes
import enum

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.rpc_wrappers.kv_write import kv_write
from infuse_iot.util.ctypes import VLACompatLittleEndianStruct
from infuse_iot.zephyr.errno import errno


class lte_pdp_ctx(InfuseRpcCommand, defs.kv_write):
    HELP = "Set the LTE PDP context (IP family & APN)"
    DESCRIPTION = "Set the LTE PDP context (IP family & APN)"

    class request(ctypes.LittleEndianStructure):
        _fields_ = [
            ("num", ctypes.c_uint8),
        ]
        _pack_ = 1

    class response(VLACompatLittleEndianStruct):
        vla_field = ("rc", 0 * ctypes.c_int16)

    class PDPFamily(enum.IntEnum):
        IPv4 = 0
        IPv6 = 1
        IPv4v6 = 2
        NonIP = 3

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--apn", "-a", type=str, required=True, help="Access Point Name")
        family_group = parser.add_mutually_exclusive_group()
        family_group.add_argument("--ipv4", dest="family", action="store_const", const=cls.PDPFamily.IPv4, help="IPv4")
        family_group.add_argument("--ipv6", dest="family", action="store_const", const=cls.PDPFamily.IPv6, help="IPv6")
        family_group.add_argument(
            "--ipv4v6",
            dest="family",
            action="store_const",
            default=cls.PDPFamily.IPv4v6,
            const=cls.PDPFamily.IPv4v6,
            help="IPv4v6",
        )
        family_group.add_argument(
            "--nonip", dest="family", action="store_const", const=cls.PDPFamily.NonIP, help="NonIP"
        )

    def __init__(self, args):
        self.args = args

    def request_struct(self):
        family_bytes = self.args.family.to_bytes(1, "little")
        apn_bytes = self.args.apn.encode("utf-8") + b"\x00"
        apn_bytes = len(apn_bytes).to_bytes(1, "little") + apn_bytes

        pdp_struct = kv_write.kv_store_value_factory(45, family_bytes + apn_bytes)
        request_bytes = bytes(pdp_struct)
        return bytes(self.request(1)) + request_bytes

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Invalid data buffer ({errno.strerror(-return_code)})")
            return

        def print_status(name, rc):
            if rc < 0:
                print(f"{name} failed to write")
            elif rc == 0:
                print(f"{name} already matched")
            else:
                print(f"{name} updated")

        print_status("PDP Context", response.rc[0])



#!/usr/bin/env python3

"""Manage RPCs through Infuse-IoT cloud"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2024, Embeint Holdings Pty Ltd"

import argparse
import base64
import importlib
import json
import pkgutil
import sys
from uuid import UUID

import infuse_iot.rpc_wrappers as wrappers
from infuse_iot.api_client import Client
from infuse_iot.api_client.api.rpc import get_rpc_by_id, send_rpc
from infuse_iot.api_client.models import Error, NewRPCMessage, NewRPCReq, RPCParams, RpcRsp
from infuse_iot.api_client.models.downlink_message_status import DownlinkMessageStatus
from infuse_iot.commands import InfuseCommand, InfuseRpcCommand
from infuse_iot.credentials import get_api_key
from infuse_iot.definitions.rpc import id_type_mapping


class SubCommand(InfuseCommand):
    NAME = "rpc_cloud"
    HELP = "Manage remote procedure calls through Infuse-IoT cloud"
    DESCRIPTION = "Manage remote procedure calls through Infuse-IoT cloud"

    @classmethod
    def add_parser(cls, parser):
        subparser = parser.add_subparsers(title="commands", metavar="<command>", required=True)

        parser_queue = subparser.add_parser("queue", help="Queue a RPC to be sent")
        parser_queue.set_defaults(_tool_action="queue")
        parser_queue.add_argument("--id", required=True, type=lambda x: int(x, 0), help="Infuse ID to run command on")
        parser_queue.add_argument("--queue-timeout", type=int, default=600, help="Timeout to send command in seconds")
        command_list_parser = parser_queue.add_subparsers(title="commands", metavar="<command>", required=True)

        for _, name, _ in pkgutil.walk_packages(wrappers.__path__):
            full_name = f"{wrappers.__name__}.{name}"
            module = importlib.import_module(full_name)

            # Add RPC wrapper to parser
            cmd_cls = getattr(module, name)
            cmd_parser = command_list_parser.add_parser(
                name,
                help=cmd_cls.HELP,
                description=cmd_cls.DESCRIPTION,
                formatter_class=argparse.RawDescriptionHelpFormatter,
            )
            cmd_parser.set_defaults(rpc_class=cmd_cls)
            cmd_cls.add_parser(cmd_parser)

        parser_query = subparser.add_parser("query", help="Query the state of a previously queued RPC")
        parser_query.set_defaults(_tool_action="query")
        parser_query.add_argument("--id", required=True, type=str, help="RPC ID from `infuse rpc_cloud queue`")

    def __init__(self, args: argparse.Namespace):
        self._args = args

    def queue(self, client: Client):
        infuse_id = f"{self._args.id:016x}"
        command: InfuseRpcCommand = self._args.rpc_class(self._args)
        timeout_ms = 1000 * self._args.queue_timeout

        assert hasattr(command, "COMMAND_ID")

        try:
            # Get the human readable arguments if implementated
            params = RPCParams.from_dict(command.request_json())
            rpc_req = NewRPCReq(command_id=command.COMMAND_ID, params=params)
        except NotImplementedError:
            # Otherwise, encode the raw binary struct
            struct_bytes = bytes(command.request_struct())
            params_encoded = base64.b64encode(struct_bytes).decode("utf-8")
            rpc_req = NewRPCReq(command_id=command.COMMAND_ID, params_encoded=params_encoded)

        rpc_msg = NewRPCMessage(infuse_id, rpc_req, timeout_ms)
        rsp = send_rpc.sync(client=client, body=rpc_msg)
        if isinstance(rsp, Error) or rsp is None:
            sys.exit(f"Failed to queue RPC ({rsp})")
        print("Query RPC state with:")
        print(f"\tinfuse rpc_cloud query --id {rsp.id}")

    def query(self, client: Client):
        rsp = get_rpc_by_id.sync(client=client, id=UUID(self._args.id))
        if isinstance(rsp, Error) or rsp is None:
            sys.exit(f"Failed to query RPC state ({rsp})")
        downlink = rsp.downlink_message
        print(f"RPC State: {downlink.status}")
        if downlink.status in [DownlinkMessageStatus.SENT, DownlinkMessageStatus.COMPLETED]:
            route = downlink.rpc_req.route
            if downlink.sent_at:
                print(f"       At: {downlink.sent_at}")
            if route:
                if route.forwarded:
                    print(f"  Through: {route.forwarded.device_id} ({route.forwarded.route.interface.upper()})")
                else:
                    print(f"  Through: Direct ({route.interface.upper()})")
        if downlink.status == DownlinkMessageStatus.COMPLETED:
            rpc_req = downlink.rpc_req
            rpc_rsp = downlink.rpc_rsp
            assert isinstance(rpc_rsp, RpcRsp)
            try:
                command_name = id_type_mapping[rpc_req.command_id].NAME
            except KeyError:
                command_name = "Unknown"
            print(f"   RPC ID: {rpc_req.command_id} ({command_name})")
            print(f"   Result: {rpc_rsp.return_code}")
            if rpc_rsp.params:
                print(json.dumps(rpc_rsp.params.additional_properties, indent=4))
            elif rpc_rsp.params_encoded:
                raw_rsp = base64.b64decode(rpc_rsp.params_encoded)
                print(f"      Raw: {raw_rsp.hex()}")
            else:
                # No response values
                pass

    def run(self):
        with Client(base_url="https://api.infuse-iot.com").with_headers(
            {"x-api-key": f"Bearer {get_api_key()}"}
        ) as client:
            if self._args._tool_action == "queue":
                self.queue(client)
            elif self._args._tool_action == "query":
                self.query(client)
            else:
                raise NotImplementedError(f"Unknown action {self._args._tool_action}")



#!/usr/bin/env python3

"""Native Bluetooth gateway tool"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2024, Embeint Holdings Pty Ltd"

import argparse
import asyncio
import ctypes
import json
from typing import Any

from bleak import BleakClient, BleakScanner
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

from infuse_iot.commands import InfuseCommand
from infuse_iot.common import InfuseBluetoothUUID, InfuseType
from infuse_iot.database import DeviceDatabase, UnknownNetworkError
from infuse_iot.epacket import interface
from infuse_iot.epacket.packet import (
    Auth,
    CtypeBtAdvFrame,
    CtypeBtGattFrame,
    Flags,
    HopReceived,
    PacketOutput,
    PacketReceived,
)
from infuse_iot.socket_comms import (
    ClientNotification,
    ClientNotificationConnectionCreated,
    ClientNotificationConnectionFailed,
    ClientNotificationEpacketReceived,
    GatewayRequest,
    GatewayRequestConnection,
    GatewayRequestConnectionRelease,
    GatewayRequestConnectionRequest,
    GatewayRequestEpacketSend,
    LocalServer,
    default_multicast_address,
)
from infuse_iot.util.argparse import BtLeAddress
from infuse_iot.util.console import Console


class InfuseGattReadResponse(ctypes.LittleEndianStructure):
    """Response to any read request on Infuse-IoT characteristics"""

    _fields_ = [
        ("cloud_public_key", 32 * ctypes.c_uint8),
        ("device_public_key", 32 * ctypes.c_uint8),
        ("network_id", ctypes.c_uint32),
    ]
    _pack_ = 1


class MulticastHandler(asyncio.DatagramProtocol):
    def __init__(self, database: DeviceDatabase, server: LocalServer, bleak_mapping: dict[int, BLEDevice]):
        self._db = database
        self._server = server
        self._mapping = bleak_mapping
        self._queues: dict[int, asyncio.Queue] = {}
        self._tasks: dict[int, asyncio.Task] = {}

    def wrapped_broadcast(self, notifcation: ClientNotification):
        try:
            self._server.broadcast(notifcation)
        except OSError as e:
            Console.log_error(f"Failed to broadcast notification: {str(e)}")

    def notification_handler(self, _characteristic: BleakGATTCharacteristic, data: bytearray):
        try:
            hdr, decr = CtypeBtGattFrame.decrypt(self._db, None, bytes(data))
        except UnknownNetworkError:
            return
        # Correct values are annoying to get here
        if_addr = interface.Address(interface.Address.BluetoothLeAddr(0, 0))
        rssi = 0

        bt_hop = HopReceived(
            hdr.device_id,
            interface.ID.BT_CENTRAL,
            if_addr,
            (Auth.DEVICE if hdr.flags & Flags.ENCR_DEVICE else Auth.NETWORK),
            hdr.key_metadata,
            hdr.gps_time,
            hdr.sequence,
            rssi,
        )
        pkt = PacketReceived(
            [bt_hop],
            hdr.type,
            bytes(decr),
        )
        Console.log_rx(pkt.ptype, len(data))
        self.wrapped_broadcast(ClientNotificationEpacketReceived(pkt))

    async def create_connection_internal(
        self, request: GatewayRequestConnectionRequest, dev: BLEDevice, queue: asyncio.Queue
    ):
        Console.log_info(f"{dev}: Initiating connection")
        async with BleakClient(dev, timeout=request.timeout_ms / 1000) as client:
            # Modified from bleak example code
            if client._backend.__class__.__name__ == "BleakClientBlueZDBus":
                await client._backend._acquire_mtu()  # type: ignore

            security_info = await client.read_gatt_char(InfuseBluetoothUUID.COMMAND_CHAR)
            resp = InfuseGattReadResponse.from_buffer_copy(security_info)
            self._db.observe_security_state(
                request.infuse_id,
                bytes(resp.cloud_public_key),
                bytes(resp.device_public_key),
                resp.network_id,
            )

            if request.data_types & request.DataType.COMMAND:
                await client.start_notify(InfuseBluetoothUUID.COMMAND_CHAR, self.notification_handler)
            if request.data_types & request.DataType.DATA:
                await client.start_notify(InfuseBluetoothUUID.DATA_CHAR, self.notification_handler)
            if request.data_types & request.DataType.LOGGING:
                await client.start_notify(InfuseBluetoothUUID.LOGGING_CHAR, self.notification_handler)

            Console.log_info(f"{dev}: Connected (MTU {client.mtu_size})")

            self.wrapped_broadcast(
                ClientNotificationConnectionCreated(
                    request.infuse_id,
                    # ATT header uses 3 bytes of the MTU
                    client.mtu_size - 3 - ctypes.sizeof(CtypeBtGattFrame) - 16,
                )
            )

            req: GatewayRequest
            while req := await queue.get():
                if isinstance(req, GatewayRequestConnectionRelease):
                    break
                assert isinstance(req, GatewayRequestEpacketSend)
                pkt: PacketOutput = req.epacket

                # Encrypt payload
                encr = CtypeBtGattFrame.encrypt(self._db, request.infuse_id, pkt.ptype, pkt.auth, pkt.payload)

                if pkt.ptype in [InfuseType.RPC_CMD, InfuseType.RPC_DATA]:
                    uuid = InfuseBluetoothUUID.COMMAND_CHAR
                else:
                    uuid = InfuseBluetoothUUID.DATA_CHAR

                Console.log_tx(pkt.ptype, len(encr))
                await client.write_gatt_char(uuid, encr, response=False)

        # Queue no longer being handled
        self._queues.pop(request.infuse_id)
        self._tasks.pop(request.infuse_id)
        Console.log_info(f"{dev}: Terminating connection")

    async def create_connection(self, request: GatewayRequestConnectionRequest, dev: BLEDevice, queue: asyncio.Queue):
        try:
            await self.create_connection_internal(request, dev, queue)
        except TimeoutError as e:
            Console.log_info(f"Timeout: {str(e)}")

    def datagram_received(self, data: bytes, addr: tuple[str | Any, int]):
        loop = asyncio.get_event_loop()
        request = GatewayRequest.from_json(json.loads(data.decode("utf-8")))

        # If not a connection request, attempt to forward to connection context
        if not isinstance(request, GatewayRequestConnectionRequest):
            if isinstance(request, GatewayRequestEpacketSend):
                queue_id = request.epacket.infuse_id
            elif isinstance(request, GatewayRequestConnection):
                queue_id = request.infuse_id
            else:
                raise RuntimeError
            q: asyncio.Queue | None = self._queues.get(queue_id, None)
            if q is not None:
                loop.call_soon(lambda: q.put_nowait(request))  # type: ignore
            return

        ble_dev = self._mapping.get(request.infuse_id, None)
        if ble_dev is None:
            self.wrapped_broadcast(ClientNotificationConnectionFailed(request.infuse_id))
            return

        # Create queue for further data transfer
        q = asyncio.Queue()
        self._queues[request.infuse_id] = q
        # Create task to handle the connection
        self._tasks[request.infuse_id] = loop.create_task(self.create_connection(request, ble_dev, q))

    def error_received(self, exc):
        Console.log_error(f"Error received: {exc}")

    def connection_lost(self, exc):
        Console.log_error("Connection closed")


class SubCommand(InfuseCommand):
    NAME = "native_bt"
    HELP = "Native Bluetooth gateway"
    DESCRIPTION = "Use the local Bluetooth adapter for Bluetooth interaction"

    @classmethod
    def add_parser(cls, parser):
        pass

    def __init__(self, args: argparse.Namespace):
        self.infuse_manu = 0x0DE4
        self.database = DeviceDatabase()
        self.server = LocalServer(default_multicast_address())
        self.bleak_mapping: dict[int, BLEDevice] = {}
        self.unknown_networks: set[int] = set()
        Console.init()

    async def server_handler(self):
        sock = self.server._input_sock
        sock.setblocking(False)
        # Wrap the socket with an asyncio Datagram protocol
        loop = asyncio.get_running_loop()
        transport, _protocol = await loop.create_datagram_endpoint(
            lambda: MulticastHandler(self.database, self.server, self.bleak_mapping),
            sock=sock,
        )
        # Keep the server running
        try:
            await asyncio.Future()  # Run forever
        finally:
            transport.close()

    def simple_callback(self, device: BLEDevice, data: AdvertisementData):
        addr = interface.Address(interface.Address.BluetoothLeAddr(0, BtLeAddress.integer_value(device.address)))
        rssi = data.rssi
        payload = data.manufacturer_data[self.infuse_manu]

        try:
            hdr, decr = CtypeBtAdvFrame.decrypt(self.database, addr.val, payload)
        except UnknownNetworkError as e:
            network_id = e.args[0]
            if network_id not in self.unknown_networks:
                self.unknown_networks.add(network_id)
                Console.log_info(f"Unknown network 0x{network_id:06x}")
            return
        self.bleak_mapping[hdr.device_id] = device

        hop = HopReceived(
            hdr.device_id,
            interface.ID.BT_ADV,
            addr,
            (Auth.DEVICE if hdr.flags & Flags.ENCR_DEVICE else Auth.NETWORK),
            hdr.key_metadata,
            hdr.gps_time,
            hdr.sequence,
            rssi,
        )

        Console.log_rx(hdr.type, len(payload))
        pkt = PacketReceived([hop], hdr.type, decr)
        notification = ClientNotificationEpacketReceived(pkt)
        try:
            self.server.broadcast(notification)
        except OSError as e:
            Console.log_error(f"Failed to broadcast notification: {str(e)}")

    async def async_bt_receiver(self):
        loop = asyncio.get_event_loop()
        handler = loop.create_task(self.server_handler())

        scanner = BleakScanner(self.simple_callback, [str(InfuseBluetoothUUID.SERVICE_UUID)], cb=dict(use_bdaddr=True))

        while True:
            Console.log_info("Starting scanner")
            async with scanner:
                await handler

    def sync_request_handler(self):
        # Loop while there are packets to send
        while req := self.server.receive():
            print(req)

    def run(self):
        asyncio.run(self.async_bt_receiver())



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.api_key_org_user_type import APIKeyOrgUserType

if TYPE_CHECKING:
    from ..models.generate_api_key_body_resource_perms import GenerateAPIKeyBodyResourcePerms


T = TypeVar("T", bound="GenerateAPIKeyBody")


@_attrs_define
class GenerateAPIKeyBody:
    """
    Attributes:
        organisation_id (UUID): ID of the organisation Example: 123e4567-e89b-12d3-a456-426614174000.
        user_type (APIKeyOrgUserType): The type of user in the organization.
        resource_perms (GenerateAPIKeyBodyResourcePerms):
    """

    organisation_id: UUID
    user_type: APIKeyOrgUserType
    resource_perms: "GenerateAPIKeyBodyResourcePerms"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        organisation_id = str(self.organisation_id)

        user_type = self.user_type.value

        resource_perms = self.resource_perms.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "organisationId": organisation_id,
                "userType": user_type,
                "resourcePerms": resource_perms,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.generate_api_key_body_resource_perms import GenerateAPIKeyBodyResourcePerms

        d = dict(src_dict)
        organisation_id = UUID(d.pop("organisationId"))

        user_type = APIKeyOrgUserType(d.pop("userType"))

        resource_perms = GenerateAPIKeyBodyResourcePerms.from_dict(d.pop("resourcePerms"))

        generate_api_key_body = cls(
            organisation_id=organisation_id,
            user_type=user_type,
            resource_perms=resource_perms,
        )

        generate_api_key_body.additional_properties = d
        return generate_api_key_body

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

T = TypeVar("T", bound="DeviceLoggerState")


@_attrs_define
class DeviceLoggerState:
    """
    Attributes:
        last_reported_block (Union[Unset, int]): Last reported block number
        last_reported_time (Union[Unset, datetime.datetime]): Last time logger state was reported
        last_downloaded_block (Union[Unset, int]): Last downloaded block number
        last_downloaded_wrap_count (Union[Unset, int]): Last downloaded block wrap count
        last_downloaded_time (Union[Unset, datetime.datetime]): Last time logger state was downloaded
    """

    last_reported_block: Unset | int = UNSET
    last_reported_time: Unset | datetime.datetime = UNSET
    last_downloaded_block: Unset | int = UNSET
    last_downloaded_wrap_count: Unset | int = UNSET
    last_downloaded_time: Unset | datetime.datetime = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        last_reported_block = self.last_reported_block

        last_reported_time: Unset | str = UNSET
        if not isinstance(self.last_reported_time, Unset):
            last_reported_time = self.last_reported_time.isoformat()

        last_downloaded_block = self.last_downloaded_block

        last_downloaded_wrap_count = self.last_downloaded_wrap_count

        last_downloaded_time: Unset | str = UNSET
        if not isinstance(self.last_downloaded_time, Unset):
            last_downloaded_time = self.last_downloaded_time.isoformat()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if last_reported_block is not UNSET:
            field_dict["lastReportedBlock"] = last_reported_block
        if last_reported_time is not UNSET:
            field_dict["lastReportedTime"] = last_reported_time
        if last_downloaded_block is not UNSET:
            field_dict["lastDownloadedBlock"] = last_downloaded_block
        if last_downloaded_wrap_count is not UNSET:
            field_dict["lastDownloadedWrapCount"] = last_downloaded_wrap_count
        if last_downloaded_time is not UNSET:
            field_dict["lastDownloadedTime"] = last_downloaded_time

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        last_reported_block = d.pop("lastReportedBlock", UNSET)

        _last_reported_time = d.pop("lastReportedTime", UNSET)
        last_reported_time: Unset | datetime.datetime
        if isinstance(_last_reported_time, Unset):
            last_reported_time = UNSET
        else:
            last_reported_time = isoparse(_last_reported_time)

        last_downloaded_block = d.pop("lastDownloadedBlock", UNSET)

        last_downloaded_wrap_count = d.pop("lastDownloadedWrapCount", UNSET)

        _last_downloaded_time = d.pop("lastDownloadedTime", UNSET)
        last_downloaded_time: Unset | datetime.datetime
        if isinstance(_last_downloaded_time, Unset):
            last_downloaded_time = UNSET
        else:
            last_downloaded_time = isoparse(_last_downloaded_time)

        device_logger_state = cls(
            last_reported_block=last_reported_block,
            last_reported_time=last_reported_time,
            last_downloaded_block=last_downloaded_block,
            last_downloaded_wrap_count=last_downloaded_wrap_count,
            last_downloaded_time=last_downloaded_time,
        )

        device_logger_state.additional_properties = d
        return device_logger_state

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3

import tabulate

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.zephyr import wifi as z_wifi
from infuse_iot.zephyr.errno import errno


class wifi_scan(InfuseRpcCommand, defs.wifi_scan):
    @classmethod
    def add_parser(cls, parser):
        return

    def __init__(self, args):
        self.args = args

    def request_struct(self):
        return self.request()

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to query current time ({errno.strerror(-return_code)})")
            return

        table = []
        for network in response.networks:
            bssid = ":".join([f"{b:02x}" for b in network.bssid])

            table.append(
                [
                    bytes(network.ssid).decode("utf-8"),
                    bssid,
                    str(z_wifi.FrequencyBand(network.band)),
                    network.channel,
                    str(z_wifi.SecurityType(network.security)),
                    f"{network.rssi} dBm",
                ]
            )

        headers = ["SSID", "BSSID", "Band", "Channel", "Security", "RSSI"]
        print(tabulate.tabulate(table, headers=headers))



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="SecurityState")


@_attrs_define
class SecurityState:
    """
    Attributes:
        cloud_public_key (str): Cloud public ECC key as a base64 encoded string
        device_public_key (str): Device public ECC key as a base64 encoded string
        network_id (int): Current network ID
        challenge (str): Challenge used in request as a base64 encoded string
        challenge_response_type (int): Type of the challenge response
        challenge_response (str): Challenge response as a base64 encoded string
    """

    cloud_public_key: str
    device_public_key: str
    network_id: int
    challenge: str
    challenge_response_type: int
    challenge_response: str
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        cloud_public_key = self.cloud_public_key

        device_public_key = self.device_public_key

        network_id = self.network_id

        challenge = self.challenge

        challenge_response_type = self.challenge_response_type

        challenge_response = self.challenge_response

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "cloudPublicKey": cloud_public_key,
                "devicePublicKey": device_public_key,
                "networkId": network_id,
                "challenge": challenge,
                "challengeResponseType": challenge_response_type,
                "challengeResponse": challenge_response,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        cloud_public_key = d.pop("cloudPublicKey")

        device_public_key = d.pop("devicePublicKey")

        network_id = d.pop("networkId")

        challenge = d.pop("challenge")

        challenge_response_type = d.pop("challengeResponseType")

        challenge_response = d.pop("challengeResponse")

        security_state = cls(
            cloud_public_key=cloud_public_key,
            device_public_key=device_public_key,
            network_id=network_id,
            challenge=challenge,
            challenge_response_type=challenge_response_type,
            challenge_response=challenge_response,
        )

        security_state.additional_properties = d
        return security_state

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.route_type import RouteType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.bt_le_route import BtLeRoute
    from ..models.forwarded_downlink_route import ForwardedDownlinkRoute
    from ..models.interface_data import InterfaceData
    from ..models.udp_downlink_route import UdpDownlinkRoute


T = TypeVar("T", bound="DownlinkRoute")


@_attrs_define
class DownlinkRoute:
    """
    Attributes:
        interface (RouteType): Interface of route
        interface_data (InterfaceData):
        udp (Union[Unset, UdpDownlinkRoute]):
        bt_adv (Union[Unset, BtLeRoute]):
        bt_peripheral (Union[Unset, BtLeRoute]):
        bt_central (Union[Unset, BtLeRoute]):
        forwarded (Union[Unset, ForwardedDownlinkRoute]):
    """

    interface: RouteType
    interface_data: "InterfaceData"
    udp: Union[Unset, "UdpDownlinkRoute"] = UNSET
    bt_adv: Union[Unset, "BtLeRoute"] = UNSET
    bt_peripheral: Union[Unset, "BtLeRoute"] = UNSET
    bt_central: Union[Unset, "BtLeRoute"] = UNSET
    forwarded: Union[Unset, "ForwardedDownlinkRoute"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        interface = self.interface.value

        interface_data = self.interface_data.to_dict()

        udp: Unset | dict[str, Any] = UNSET
        if not isinstance(self.udp, Unset):
            udp = self.udp.to_dict()

        bt_adv: Unset | dict[str, Any] = UNSET
        if not isinstance(self.bt_adv, Unset):
            bt_adv = self.bt_adv.to_dict()

        bt_peripheral: Unset | dict[str, Any] = UNSET
        if not isinstance(self.bt_peripheral, Unset):
            bt_peripheral = self.bt_peripheral.to_dict()

        bt_central: Unset | dict[str, Any] = UNSET
        if not isinstance(self.bt_central, Unset):
            bt_central = self.bt_central.to_dict()

        forwarded: Unset | dict[str, Any] = UNSET
        if not isinstance(self.forwarded, Unset):
            forwarded = self.forwarded.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "interface": interface,
                "interfaceData": interface_data,
            }
        )
        if udp is not UNSET:
            field_dict["udp"] = udp
        if bt_adv is not UNSET:
            field_dict["btAdv"] = bt_adv
        if bt_peripheral is not UNSET:
            field_dict["btPeripheral"] = bt_peripheral
        if bt_central is not UNSET:
            field_dict["btCentral"] = bt_central
        if forwarded is not UNSET:
            field_dict["forwarded"] = forwarded

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.bt_le_route import BtLeRoute
        from ..models.forwarded_downlink_route import ForwardedDownlinkRoute
        from ..models.interface_data import InterfaceData
        from ..models.udp_downlink_route import UdpDownlinkRoute

        d = dict(src_dict)
        interface = RouteType(d.pop("interface"))

        interface_data = InterfaceData.from_dict(d.pop("interfaceData"))

        _udp = d.pop("udp", UNSET)
        udp: Unset | UdpDownlinkRoute
        if isinstance(_udp, Unset):
            udp = UNSET
        else:
            udp = UdpDownlinkRoute.from_dict(_udp)

        _bt_adv = d.pop("btAdv", UNSET)
        bt_adv: Unset | BtLeRoute
        if isinstance(_bt_adv, Unset):
            bt_adv = UNSET
        else:
            bt_adv = BtLeRoute.from_dict(_bt_adv)

        _bt_peripheral = d.pop("btPeripheral", UNSET)
        bt_peripheral: Unset | BtLeRoute
        if isinstance(_bt_peripheral, Unset):
            bt_peripheral = UNSET
        else:
            bt_peripheral = BtLeRoute.from_dict(_bt_peripheral)

        _bt_central = d.pop("btCentral", UNSET)
        bt_central: Unset | BtLeRoute
        if isinstance(_bt_central, Unset):
            bt_central = UNSET
        else:
            bt_central = BtLeRoute.from_dict(_bt_central)

        _forwarded = d.pop("forwarded", UNSET)
        forwarded: Unset | ForwardedDownlinkRoute
        if isinstance(_forwarded, Unset):
            forwarded = UNSET
        else:
            forwarded = ForwardedDownlinkRoute.from_dict(_forwarded)

        downlink_route = cls(
            interface=interface,
            interface_data=interface_data,
            udp=udp,
            bt_adv=bt_adv,
            bt_peripheral=bt_peripheral,
            bt_central=bt_central,
            forwarded=forwarded,
        )

        downlink_route.additional_properties = d
        return downlink_route

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3

import time
from abc import ABCMeta, abstractmethod
from io import BufferedWriter

import pylink
import serial
from pyocd.core.helpers import ConnectHelper
from pyocd.debug.rtt import GenericRTTControlBlock


class SerialFrame:
    """Serial frame reconstructor"""

    SYNC = b"\xd5\xca"

    @classmethod
    def reconstructor(cls):
        length = 0
        buffered = bytearray()
        packet = None
        while True:
            # Is the current byte part of a packet?
            consumed = len(buffered) > 0 or packet is not None
            # Get next byte from port, yield current state
            val = yield consumed, packet
            packet = None
            # Append value to buffer
            buffered.append(val)
            # Wait for packet sync bytes
            if len(buffered) <= 2:
                if val != cls.SYNC[len(buffered) - 1]:
                    buffered = bytearray()
                    continue
            # Store length
            elif len(buffered) == 4:
                length = int.from_bytes(buffered[2:], "little")
            # Complete packet received
            elif len(buffered) == 4 + length:
                packet = buffered[4:]
                buffered = bytearray()


class SerialLike(metaclass=ABCMeta):
    @abstractmethod
    def open(self, timeout: float | None = None) -> None:
        """Open serial port"""

    @abstractmethod
    def read_bytes(self, num: int):
        """Read arbitrary number of bytes from serial port"""

    @abstractmethod
    def ping(self):
        """Magic 1 byte frame to request a response"""

    @abstractmethod
    def write(self, packet: bytes):
        """Write a serial frame to the port"""

    @abstractmethod
    def close(self):
        """Close the serial port"""

    @abstractmethod
    def __str__(self):
        """Human readable description of the port"""


class SerialPort(SerialLike):
    """Serial Port handling"""

    def __init__(self, serial_port, baudrate=115200):
        self._ser = serial.Serial()
        self._ser.port = str(serial_port)
        self._ser.baudrate = baudrate
        self._ser.timeout = 0.05
        # Prepend leading 0's for high baudrates to give sleepy
        # receivers (STM32) time to wake up on RX before real data arrives.
        self._prefix = b"\x00\x00" if baudrate > 115200 else b""

    def open(self, timeout: float | None = None):
        self._ser.open()

    def read_bytes(self, num) -> bytes:
        return self._ser.read(num)

    def ping(self) -> None:
        self._ser.write(self._prefix + SerialFrame.SYNC + b"\x01\x00" + b"\x4d")
        self._ser.flush()

    def write(self, packet: bytes) -> None:
        # Add header
        pkt = self._prefix + SerialFrame.SYNC + len(packet).to_bytes(2, "little") + packet
        # Write packet to serial port
        self._ser.write(pkt)
        self._ser.flush()

    def close(self) -> None:
        self._ser.close()

    def __str__(self) -> str:
        if self._ser.port:
            return f"Serial: {self._ser.port}"
        return "Serial: N/A"


class RttPort(SerialLike):
    """Segger RTT handling"""

    def __init__(self, rtt_device: str, serial_number: str | None = None):
        self._jlink = pylink.JLink()
        self._name = rtt_device
        self._serial_number = serial_number
        self._modem_trace: BufferedWriter | None = None
        self._modem_trace_buf = 0

    def open(self, timeout: float | None = None):
        self._jlink.open(serial_no=self._serial_number)
        self._jlink.set_tif(pylink.enums.JLinkInterfaces.SWD)
        self._jlink.connect(self._name, 4000)
        self._jlink.rtt_start()

        end_time = time.time() + timeout if timeout else None

        # Loop until JLink initialised properly
        while True:
            if end_time and time.time() > end_time:
                raise TimeoutError("RTT port never initialised")
            try:
                num_up = self._jlink.rtt_get_num_up_buffers()
                _num_down = self._jlink.rtt_get_num_down_buffers()
                break
            except pylink.errors.JLinkRTTException:
                time.sleep(0.05)

        # Do a search and see if we have a modem trace file
        for i in range(num_up):
            desc = self._jlink.rtt_get_buf_descriptor(i, True)
            if desc.name == "modem_trace":
                f = f"{int(time.time())}_nrf_modem_trace.bin"
                print(f"Found nRF LTE modem trace channel (opening {f:s})")
                self._modem_trace = open(f, mode="wb")  # noqa: SIM115
                self._modem_trace_buf = desc.BufferIndex

    def read_bytes(self, num):
        if self._modem_trace is not None:
            trace_data = bytes(self._jlink.rtt_read(self._modem_trace_buf, 1024))
            if len(trace_data) > 0:
                self._modem_trace.write(trace_data)

        return bytes(self._jlink.rtt_read(0, num))

    def ping(self):
        self._jlink.rtt_write(0, SerialFrame.SYNC + b"\x01\x00" + b"\x4d")

    def write(self, packet: bytes):
        # Add header
        pkt = SerialFrame.SYNC + len(packet).to_bytes(2, "little") + packet
        while True:
            res = self._jlink.rtt_write(0, pkt)
            if res == len(pkt):
                break
            pkt = pkt[res:]
            time.sleep(0.1)

    def close(self):
        self._jlink.rtt_stop()
        self._jlink.close()
        if self._modem_trace is not None:
            self._modem_trace.flush()
            self._modem_trace.close()

    def __str__(self) -> str:
        return f"RTT: {self._name}"


class PyOcdPort(SerialLike):
    """PyOcd RTT handling"""

    def __init__(self, target: str):
        self._name = target
        self._session = ConnectHelper.session_with_chosen_probe(auto_open=False, options={"target_override": target})
        self._target = self._session.target
        self._rtt = GenericRTTControlBlock(self._target)

    def open(self, timeout: float | None = None):
        self._session.open()
        self._target.resume()
        self._rtt.start()
        self._up_chan = self._rtt.up_channels[0]
        self._down_chan = self._rtt.down_channels[0]

    def read_bytes(self, num):
        return self._up_chan.read()

    def ping(self):
        self._down_chan.write(SerialFrame.SYNC + b"\x01\x00" + b"\x4d")

    def write(self, packet: bytes):
        # Add header
        pkt = SerialFrame.SYNC + len(packet).to_bytes(2, "little") + packet
        while True:
            res = self._down_chan.write(pkt)
            if res == len(pkt):
                break
            pkt = pkt[res:]
            time.sleep(0.1)

    def close(self):
        self._session.close()

    def __str__(self) -> str:
        return f"PYOCD: {self._name}"



#!/usr/bin/env python3

import argparse
import pathlib
import re
from typing import cast

import yaml

from infuse_iot.definitions.rpc import rpc_enum_bt_le_addr_type, rpc_struct_bt_addr_le
from infuse_iot.util.ctypes import bytes_to_uint8


class ValidFile:
    """Filesystem path that exists"""

    def __new__(cls, string) -> pathlib.Path:  # type: ignore
        p = pathlib.Path(string)
        if p.exists():
            return p
        else:
            raise argparse.ArgumentTypeError(f"{string} does not exist")


class ValidDir:
    """Filesystem directory that exists"""

    def __new__(cls, string) -> pathlib.Path:  # type: ignore
        p = pathlib.Path(string)
        if not p.exists():
            raise argparse.ArgumentTypeError(f"{string} does not exist")
        if not p.is_dir():
            raise argparse.ArgumentTypeError(f"{string} is not a directory")
        return p


class ValidRelease:
    """Infuse-IoT release folder"""

    def __init__(self, string):
        p: pathlib.Path = ValidDir(string)  # type: ignore
        metadata = p / "manifest.yaml"
        if not metadata.exists():
            raise argparse.ArgumentTypeError(f"{string} is not an Infuse-IoT release")
        self.dir = p
        metadata = self.dir / "manifest.yaml"
        with open(metadata, encoding="utf-8") as f:
            self.metadata = yaml.safe_load(f.read())


class BtLeAddress:
    """Bluetooth Low-Energy address"""

    def __new__(cls, string) -> int:  # type: ignore
        pattern = r"((([0-9a-fA-F]{2}):){5})([0-9a-fA-F]{2})"

        if re.match(pattern, string):
            mac_cleaned = string.replace(":", "").replace("-", "")
            addr = int(mac_cleaned, 16)
        else:
            try:
                addr = int(string, 16)
            except ValueError:
                raise argparse.ArgumentTypeError(f"{string} is not a Bluetooth address") from None
        return addr

    @classmethod
    def to_ctype(cls, addr_type: rpc_enum_bt_le_addr_type, value: int) -> rpc_struct_bt_addr_le:
        """Get a ctype representation of the Bluetooth address"""
        return rpc_struct_bt_addr_le(
            addr_type,
            bytes_to_uint8(value.to_bytes(6, "little")),
        )

    @classmethod
    def integer_value(cls, string) -> int:
        """Integer value from address string"""
        return cast(int, cls(string))



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="DeviceIdField")


@_attrs_define
class DeviceIdField:
    """
    Attributes:
        device_id (Union[Unset, str]): 8 byte DeviceID as a hex string (if not provided will be auto-generated) Example:
            d291d4d66bf0a955.
    """

    device_id: Unset | str = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        device_id = self.device_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if device_id is not UNSET:
            field_dict["deviceId"] = device_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        device_id = d.pop("deviceId", UNSET)

        device_id_field = cls(
            device_id=device_id,
        )

        device_id_field.additional_properties = d
        return device_id_field

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3


import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.zephyr.errno import errno


class application_info(InfuseRpcCommand, defs.application_info):
    @classmethod
    def add_parser(cls, _parser):
        pass

    def __init__(self, _args):
        pass

    def request_struct(self):
        return self.request()

    def request_json(self):
        return {}

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to query current time ({errno.strerror(-return_code)})")
            return

        r = response
        v = r.version
        print(f"\tApplication: 0x{r.application_id:08x}")
        print(f"\t    Version: {v.major}.{v.minor}.{v.revision}+{v.build_num:08x}")
        print(f"\t    Network: 0x{r.network_id:08x}")
        print(f"\t     Uptime: {r.uptime}")
        print(f"\t    Reboots: {r.reboots}")
        print(f"\t     KV CRC: 0x{r.kv_crc:08x}")
        print(f"\t   O Blocks: {r.data_blocks_internal}")
        print(f"\t   E Blocks: {r.data_blocks_external}")



"""Contains all the data models used in inputs/outputs"""

from .algorithm import Algorithm
from .api_key_org_user_type import APIKeyOrgUserType
from .api_key_resource_name import APIKeyResourceName
from .api_key_resource_perm import APIKeyResourcePerm
from .application_version import ApplicationVersion
from .board import Board
from .bt_le_route import BtLeRoute
from .bt_le_route_type import BtLeRouteType
from .coap_file_stats import COAPFileStats
from .coap_files_list import COAPFilesList
from .created_board_properties import CreatedBoardProperties
from .created_device_properties import CreatedDeviceProperties
from .created_organisation_properties import CreatedOrganisationProperties
from .created_rpc_message import CreatedRpcMessage
from .definitions_enum_definition import DefinitionsEnumDefinition
from .definitions_enum_value import DefinitionsEnumValue
from .definitions_field_conversion import DefinitionsFieldConversion
from .definitions_field_conversion_int import DefinitionsFieldConversionInt
from .definitions_field_definition import DefinitionsFieldDefinition
from .definitions_field_display import DefinitionsFieldDisplay
from .definitions_field_display_fmt import DefinitionsFieldDisplayFmt
from .definitions_kv import DefinitionsKV
from .definitions_kv_definition import DefinitionsKVDefinition
from .definitions_kv_definitions import DefinitionsKVDefinitions
from .definitions_kv_response import DefinitionsKVResponse
from .definitions_kv_structs import DefinitionsKVStructs
from .definitions_rpc import DefinitionsRPC
from .definitions_rpc_command import DefinitionsRPCCommand
from .definitions_rpc_command_default_auth import DefinitionsRPCCommandDefaultAuth
from .definitions_rpc_commands import DefinitionsRPCCommands
from .definitions_rpc_enums import DefinitionsRPCEnums
from .definitions_rpc_response import DefinitionsRPCResponse
from .definitions_rpc_structs import DefinitionsRPCStructs
from .definitions_struct_definition import DefinitionsStructDefinition
from .definitions_tdf import DefinitionsTDF
from .definitions_tdf_definition import DefinitionsTDFDefinition
from .definitions_tdf_definitions import DefinitionsTDFDefinitions
from .definitions_tdf_response import DefinitionsTDFResponse
from .definitions_tdf_structs import DefinitionsTDFStructs
from .derive_device_key_body import DeriveDeviceKeyBody
from .device import Device
from .device_and_state import DeviceAndState
from .device_entry_update_status import DeviceEntryUpdateStatus
from .device_id_field import DeviceIdField
from .device_kv_entry import DeviceKVEntry
from .device_kv_entry_decoded import DeviceKVEntryDecoded
from .device_kv_entry_update import DeviceKVEntryUpdate
from .device_logger_state import DeviceLoggerState
from .device_metadata import DeviceMetadata
from .device_metadata_update import DeviceMetadataUpdate
from .device_metadata_update_operation import DeviceMetadataUpdateOperation
from .device_state import DeviceState
from .device_update import DeviceUpdate
from .downlink_message import DownlinkMessage
from .downlink_message_status import DownlinkMessageStatus
from .downlink_route import DownlinkRoute
from .error import Error
from .forwarded_downlink_route import ForwardedDownlinkRoute
from .forwarded_uplink_route import ForwardedUplinkRoute
from .generate_api_key_body import GenerateAPIKeyBody
from .generate_api_key_body_resource_perms import GenerateAPIKeyBodyResourcePerms
from .generate_mqtt_token_body import GenerateMQTTTokenBody
from .generated_api_key import GeneratedAPIKey
from .generated_mqtt_token import GeneratedMQTTToken
from .get_last_routes_for_devices_body import GetLastRoutesForDevicesBody
from .health_check import HealthCheck
from .interface_data import InterfaceData
from .key import Key
from .key_interface import KeyInterface
from .metadata_field import MetadataField
from .new_board import NewBoard
from .new_device import NewDevice
from .new_device_kv_entry_update import NewDeviceKVEntryUpdate
from .new_device_kv_entry_update_decoded import NewDeviceKVEntryUpdateDecoded
from .new_device_state import NewDeviceState
from .new_organisation import NewOrganisation
from .new_rpc_message import NewRPCMessage
from .new_rpc_req import NewRPCReq
from .organisation import Organisation
from .route_type import RouteType
from .rpc_message import RpcMessage
from .rpc_params import RPCParams
from .rpc_req import RpcReq
from .rpc_req_data_header import RPCReqDataHeader
from .rpc_rsp import RpcRsp
from .security_state import SecurityState
from .udp_downlink_route import UdpDownlinkRoute
from .udp_uplink_route import UdpUplinkRoute
from .uplink_route import UplinkRoute
from .uplink_route_and_device_id import UplinkRouteAndDeviceId

__all__ = (
    "Algorithm",
    "APIKeyOrgUserType",
    "APIKeyResourceName",
    "APIKeyResourcePerm",
    "ApplicationVersion",
    "Board",
    "BtLeRoute",
    "BtLeRouteType",
    "COAPFilesList",
    "COAPFileStats",
    "CreatedBoardProperties",
    "CreatedDeviceProperties",
    "CreatedOrganisationProperties",
    "CreatedRpcMessage",
    "DefinitionsEnumDefinition",
    "DefinitionsEnumValue",
    "DefinitionsFieldConversion",
    "DefinitionsFieldConversionInt",
    "DefinitionsFieldDefinition",
    "DefinitionsFieldDisplay",
    "DefinitionsFieldDisplayFmt",
    "DefinitionsKV",
    "DefinitionsKVDefinition",
    "DefinitionsKVDefinitions",
    "DefinitionsKVResponse",
    "DefinitionsKVStructs",
    "DefinitionsRPC",
    "DefinitionsRPCCommand",
    "DefinitionsRPCCommandDefaultAuth",
    "DefinitionsRPCCommands",
    "DefinitionsRPCEnums",
    "DefinitionsRPCResponse",
    "DefinitionsRPCStructs",
    "DefinitionsStructDefinition",
    "DefinitionsTDF",
    "DefinitionsTDFDefinition",
    "DefinitionsTDFDefinitions",
    "DefinitionsTDFResponse",
    "DefinitionsTDFStructs",
    "DeriveDeviceKeyBody",
    "Device",
    "DeviceAndState",
    "DeviceEntryUpdateStatus",
    "DeviceIdField",
    "DeviceKVEntry",
    "DeviceKVEntryDecoded",
    "DeviceKVEntryUpdate",
    "DeviceLoggerState",
    "DeviceMetadata",
    "DeviceMetadataUpdate",
    "DeviceMetadataUpdateOperation",
    "DeviceState",
    "DeviceUpdate",
    "DownlinkMessage",
    "DownlinkMessageStatus",
    "DownlinkRoute",
    "Error",
    "ForwardedDownlinkRoute",
    "ForwardedUplinkRoute",
    "GenerateAPIKeyBody",
    "GenerateAPIKeyBodyResourcePerms",
    "GeneratedAPIKey",
    "GeneratedMQTTToken",
    "GenerateMQTTTokenBody",
    "GetLastRoutesForDevicesBody",
    "HealthCheck",
    "InterfaceData",
    "Key",
    "KeyInterface",
    "MetadataField",
    "NewBoard",
    "NewDevice",
    "NewDeviceKVEntryUpdate",
    "NewDeviceKVEntryUpdateDecoded",
    "NewDeviceState",
    "NewOrganisation",
    "NewRPCMessage",
    "NewRPCReq",
    "Organisation",
    "RouteType",
    "RpcMessage",
    "RPCParams",
    "RpcReq",
    "RPCReqDataHeader",
    "RpcRsp",
    "SecurityState",
    "UdpDownlinkRoute",
    "UdpUplinkRoute",
    "UplinkRoute",
    "UplinkRouteAndDeviceId",
)



#!/usr/bin/env python3


import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.zephyr.errno import errno


class fault(InfuseRpcCommand, defs.fault):
    @classmethod
    def add_parser(cls, parser):
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument(
            "--div-0",
            dest="fault",
            action="store_const",
            const=30,
            help="Divide by zero exception",
        )
        group.add_argument(
            "--null-deref",
            dest="fault",
            action="store_const",
            const=19,
            help="NULL dereference exception",
        )
        group.add_argument(
            "--stack-overflow",
            dest="fault",
            action="store_const",
            const=2,
            help="Stack overflow exception",
        )
        group.add_argument(
            "--assert",
            dest="fault",
            action="store_const",
            const=3,
            help="Assertion failure",
        )
        group.add_argument(
            "--undef-instr",
            dest="fault",
            action="store_const",
            const=36,
            help="Undefined instruction exception",
        )
        group.add_argument(
            "--instr-access",
            dest="fault",
            action="store_const",
            const=20,
            help="Instruction access exception",
        )
        group.add_argument(
            "--wdog",
            dest="fault",
            action="store_const",
            const=128,
            help="Watchdog timeout exception",
        )
        group.add_argument(
            "--secure",
            dest="fault",
            action="store_const",
            const=37,
            help="Secure code exception",
        )

    def __init__(self, args):
        self._fault_type = args.fault

    def request_struct(self):
        return self.request(self._fault_type, 0)

    def handle_response(self, return_code, _):
        print(f"Failed to trigger exception ({errno.strerror(-return_code)})")



#!/usr/bin/env python3

import tabulate

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.zephyr.errno import errno


class infuse_states_query(InfuseRpcCommand, defs.infuse_states_query):
    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--offset", type=int, default=0)

    def __init__(self, args):
        self.args = args

    def request_struct(self):
        return self.request(self.args.offset)

    def request_json(self):
        return {}

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to query states ({errno.strerror(-return_code)})")
            return

        states = []
        for state in response.states:
            states.append([state.state, "Permanent" if state.timeout == 0 else f"{state.timeout} seconds"])
        print(tabulate.tabulate(states, headers=["State", "Duration"]))



#!/usr/bin/env python3

import binascii
import sys

import tabulate
from elftools.dwarf.die import DIE
from elftools.elf.elffile import ELFFile

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.util import console, elftools
from infuse_iot.util.argparse import ValidFile
from infuse_iot.zephyr.errno import errno


class sym_read(InfuseRpcCommand, defs.mem_read):
    RPC_DATA_RECEIVE = True

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--elf", type=ValidFile, help="ELF file to read symbol data from")
        read_type = parser.add_mutually_exclusive_group(required=True)
        read_type.add_argument("--sym", type=str, help="Symbol name to read")
        read_type.add_argument("--addr", type=lambda x: int(x, 0), help="Address to read")

    def __init__(self, args):
        # Ignore context-manager warning since ELFFile requires the file to remain opened
        self.elf_file = open(args.elf, "rb")  # noqa: SIM115
        self.elf = ELFFile(self.elf_file)
        self.symbol_die: DIE | None

        if args.sym:
            symbols = elftools.symbols_from_name(self.elf, args.sym)
            if len(symbols) == 0:
                sys.exit(f"{args.sym} not found in '{args.elf}' symbol table")
            elif len(symbols) == 1:
                self.symbol_die = elftools.dwarf_die_from_symbol(self.elf, symbols[0])
                idx = 0
            else:
                dies = []
                options = []
                # User readable selection requires the filename and line number
                for s in symbols:
                    die = elftools.dwarf_die_from_symbol(self.elf, s)
                    if die is None:
                        continue
                    filename, linenum = elftools.dwarf_die_file_info(self.elf, die)
                    dies.append(die)
                    options.append(f"{filename}:{linenum}")
                # Ask the user which symbol they mean
                try:
                    idx, _ = console.choose_one(f"Multiple symbols matching '{args.sym}', choose one:", options)
                except IndexError:
                    sys.exit("No symbol chosen...")
                self.symbol_die = dies[idx]

            self.symbol = symbols[idx]
        elif args.addr:
            symbol = elftools.symbol_from_address(self.elf, args.addr)
            if symbol is None:
                sys.exit(f"Could not find symbol for address 0x{args.addr:08x} in '{args.elf}' symbol table")
            self.symbol = symbol
        else:
            raise NotImplementedError("Unexpected symbol refrence")

        self.address = self.symbol.entry["st_value"]
        self.num = self.symbol.entry["st_size"]

        dwarf_info = self.elf.get_dwarf_info()
        if self.symbol_die is not None:
            self.symbol_info = elftools.dwarf_die_variable_inf(dwarf_info, self.symbol_die)

        self.expected_offset = 0
        self.output = b""

    def request_struct(self):
        return self.request(self.address)

    def data_payload_recv_len(self):
        return self.num

    def data_recv_cb(self, offset: int, data: bytes) -> None:
        if offset != self.expected_offset:
            missing = offset - self.expected_offset
            print(f"Missed {missing:d} bytes from offset 0x{self.expected_offset:08x}")
            self.output += b"\x00" * missing

        self.output += data
        # Next expected offset
        self.expected_offset = offset + len(data)

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Failed to read data logger ({errno.strerror(-return_code)})")
            return

        if response.sent_len != len(self.output):
            print(f"Unexpected received length ({response.sent_len} != {len(self.output)})")
            return

        if response.sent_crc != binascii.crc32(self.output):
            print(f"Unexpected received length ({response.sent_crc:08x} != {binascii.crc32(self.output)}:08x)")
            return

        if self.elf is None:
            # Hexdump the received payload
            for offset in range(0, len(self.output), 16):
                print(f"{self.address + offset:08x}: {self.output[offset : offset + 16].hex()}")
            return

        # Parse returned value
        symbol_size = self.symbol.entry["st_size"]
        assert len(self.output) >= symbol_size

        if self.symbol_die is not None:
            filename, linenum = elftools.dwarf_die_file_info(self.elf, self.symbol_die)
            print(f" Symbol: {self.symbol.name} ({filename}:{linenum})")

        address_base = self.symbol.entry["st_value"]
        print(f"Address: 0x{address_base:x}")
        print(f"   Size: {symbol_size} bytes")
        if symbol_size <= 32:
            print(f"    Raw: {self.output.hex()}")
        else:
            print(f"    Raw: {self.output[:32].hex()}...")

        def info_table(info, offset=0):
            table = [[f"{'  ' * offset}{info.name}", f"({info.tag}) ({info.ctype}) {info.offset}", ""]]
            for child in info.children:
                table += info_table(child, offset + 1)
            return table

        def field_table(info: elftools.dwarf_field, buffer: bytes, offset: int = 0):
            if info.ctype is None:
                table = [[f"0x{address_base + info.offset:08x}", f"{'  ' * offset}{info.name}", "", "", ""]]
            else:
                value = info.ctype.from_buffer_copy(buffer, info.offset).value
                value_hex = hex(value) if not isinstance(value, float) else "N/A"
                points_to = ""
                if info.tag == "DW_TAG_pointer_type":
                    if value == 0x00:
                        points_to = "NULL"
                    else:
                        sym = elftools.symbol_from_address(self.elf, value)
                        if sym:
                            ptr_offset = ""
                            if value != sym.entry["st_value"]:
                                ptr_offset = f" (+ {value - sym.entry['st_value']})"
                            points_to = f"{sym.name}{ptr_offset}"
                        else:
                            points_to = "<unknown>"
                table = [
                    [f"0x{address_base + info.offset:08x}", f"{'  ' * offset}{info.name}", value, value_hex, points_to]
                ]
            for child in info.children:
                table += field_table(child, buffer, offset + 1)
            return table

        tabulate.PRESERVE_WHITESPACE = True
        print(
            tabulate.tabulate(field_table(self.symbol_info, self.output), ["Address", "Field", "Value", "Hex", "Ptr"])
        )



#!/usr/bin/env python3

import ctypes


class RequestHeader(ctypes.LittleEndianStructure):
    """RPC_CMD packet header"""

    _fields_ = [
        ("request_id", ctypes.c_uint32),
        ("command_id", ctypes.c_uint16),
    ]
    _pack_ = 1


class RequestDataHeader(ctypes.LittleEndianStructure):
    """RPC_CMD additional header for RPC_DATA"""

    _fields_ = [
        ("size", ctypes.c_uint32),
        ("rx_ack_period", ctypes.c_uint8),
    ]
    _pack_ = 1


class DataHeader(ctypes.LittleEndianStructure):
    """RPC_DATA header"""

    _fields_ = [
        ("request_id", ctypes.c_uint32),
        ("offset", ctypes.c_uint32),
    ]
    _pack_ = 1


class DataAck(ctypes.LittleEndianStructure):
    """RPC_DATA_ACK payload"""

    _fields_ = [
        ("request_id", ctypes.c_uint32),
        ("offset", 0 * ctypes.c_uint32),
    ]
    _pack_ = 1


class ResponseHeader(ctypes.LittleEndianStructure):
    """RPC_RSP packet header"""

    _fields_ = [
        ("request_id", ctypes.c_uint32),
        ("command_id", ctypes.c_uint16),
        ("return_code", ctypes.c_int16),
    ]
    _pack_ = 1



#!/usr/bin/env python3
# mypy: ignore-errors

"""Autogenerated KV store definitions"""

import ctypes

from infuse_iot.util.ctypes import VLACompatLittleEndianStruct


class structs:
    class gcs_location(VLACompatLittleEndianStruct):
        """Geographic Co-ordinate System location"""

        _fields_ = [
            ("latitude", ctypes.c_int32),
            ("longitude", ctypes.c_int32),
            ("height", ctypes.c_int32),
        ]
        _pack_ = 1

    class bt_addr_le(VLACompatLittleEndianStruct):
        """Bluetooth address type (bt_addr_le_t)"""

        _fields_ = [
            ("type", ctypes.c_uint8),
            ("val", 6 * ctypes.c_uint8),
        ]
        _pack_ = 1

    class kv_string(VLACompatLittleEndianStruct):
        """String type"""

        _fields_ = [
            ("value_num", ctypes.c_uint8),
        ]
        vla_field = ("value", 0 * ctypes.c_char)
        _pack_ = 1

        def __str__(self) -> str:
            return bytes(self.value).decode("utf-8")

    class kv_mcuboot_img_sem_ver(VLACompatLittleEndianStruct):
        """MCUboot semantic versioning struct"""

        _fields_ = [
            ("major", ctypes.c_uint8),
            ("minor", ctypes.c_uint8),
            ("revision", ctypes.c_uint16),
            ("build_num", ctypes.c_uint32),
        ]
        _pack_ = 1

    class kv_range_u8(VLACompatLittleEndianStruct):
        """Generic range structure"""

        _fields_ = [
            ("lower", ctypes.c_uint8),
            ("upper", ctypes.c_uint8),
        ]
        _pack_ = 1

    class kv_utc_hms(VLACompatLittleEndianStruct):
        """UTC Hour-Minute-Second"""

        _fields_ = [
            ("hour", ctypes.c_uint8),
            ("minute", ctypes.c_uint8),
            ("second", ctypes.c_uint8),
        ]
        _pack_ = 1

    class kv_algorithm_logging(VLACompatLittleEndianStruct):
        """Algorithm logging configuration"""

        _fields_ = [
            ("loggers", ctypes.c_uint8),
            ("tdf_mask", ctypes.c_uint8),
        ]
        _pack_ = 1

    class kv_algorithm_stationary_windowed_args(VLACompatLittleEndianStruct):
        """Arguments for 'Stationary Windowed' algorithm"""

        _fields_ = [
            ("window_seconds", ctypes.c_uint32),
            ("std_dev_threshold_ug", ctypes.c_uint32),
        ]
        _pack_ = 1

    class kv_algorithm_tilt_args(VLACompatLittleEndianStruct):
        """Arguments for 'Tilt' algorithm"""

        _fields_ = [
            ("iir_filter_alpha", ctypes.c_float),
            ("one_g_percent", ctypes.c_uint8),
        ]
        _pack_ = 1


class slots:
    class reboots(VLACompatLittleEndianStruct):
        """Number of times device has rebooted"""

        NAME = "REBOOTS"
        BASE_ID = 0
        RANGE = 1
        _fields_ = [
            ("count", ctypes.c_uint32),
        ]
        _pack_ = 1

    class bluetooth_addr(VLACompatLittleEndianStruct):
        """Bluetooth advertising address"""

        NAME = "BLUETOOTH_ADDR"
        BASE_ID = 1
        RANGE = 1
        _fields_ = [
            ("address", structs.bt_addr_le),
        ]
        _pack_ = 1

    class exfat_disk_info(VLACompatLittleEndianStruct):
        """exFAT disk information"""

        NAME = "EXFAT_DISK_INFO"
        BASE_ID = 2
        RANGE = 1
        _fields_ = [
            ("block_count", ctypes.c_uint32),
            ("block_size", ctypes.c_uint32),
        ]
        _pack_ = 1

    class bluetooth_ctlr_version(VLACompatLittleEndianStruct):
        """External Bluetooth controller version"""

        NAME = "BLUETOOTH_CTLR_VERSION"
        BASE_ID = 3
        RANGE = 1
        _fields_ = [
            ("application", ctypes.c_uint32),
            ("version", structs.kv_mcuboot_img_sem_ver),
        ]
        _pack_ = 1

    class device_name(VLACompatLittleEndianStruct):
        """Personalised name for the device"""

        NAME = "DEVICE_NAME"
        BASE_ID = 4
        RANGE = 1
        _fields_ = []
        vla_field = ("name", structs.kv_string)
        _pack_ = 1

    class infuse_application_id(VLACompatLittleEndianStruct):
        """CONFIG_INFUSE_APPLICATION_ID, store will be reset if the values don't match"""

        NAME = "INFUSE_APPLICATION_ID"
        BASE_ID = 5
        RANGE = 1
        _fields_ = [
            ("application_id", ctypes.c_uint32),
        ]
        _pack_ = 1

    class application_active(VLACompatLittleEndianStruct):
        """Control STATE_APPLICATION_ACTIVE"""

        NAME = "APPLICATION_ACTIVE"
        BASE_ID = 6
        RANGE = 1
        _fields_ = [
            ("active", ctypes.c_uint8),
        ]
        _pack_ = 1

    class board_target(VLACompatLittleEndianStruct):
        """Value of CONFIG_BOARD_TARGET"""

        NAME = "BOARD_TARGET"
        BASE_ID = 7
        RANGE = 1
        _fields_ = []
        vla_field = ("board_target", structs.kv_string)
        _pack_ = 1

    class fixed_location(VLACompatLittleEndianStruct):
        """Fixed global location of the device"""

        NAME = "FIXED_LOCATION"
        BASE_ID = 10
        RANGE = 1
        _fields_ = [
            ("location", structs.gcs_location),
        ]
        _pack_ = 1

    class wifi_ssid(VLACompatLittleEndianStruct):
        """WiFi network name"""

        NAME = "WIFI_SSID"
        BASE_ID = 20
        RANGE = 1
        _fields_ = []
        vla_field = ("ssid", structs.kv_string)
        _pack_ = 1

    class wifi_psk(VLACompatLittleEndianStruct):
        """WiFi network password"""

        NAME = "WIFI_PSK"
        BASE_ID = 21
        RANGE = 1
        _fields_ = []
        vla_field = ("psk", structs.kv_string)
        _pack_ = 1

    class wifi_channels(VLACompatLittleEndianStruct):
        """WiFi band and channel configuration"""

        NAME = "WIFI_CHANNELS"
        BASE_ID = 22
        RANGE = 1
        _fields_ = [
            ("band", ctypes.c_uint8),
        ]
        vla_field = ("channels", 0 * ctypes.c_uint8)
        _pack_ = 1

    class ntp_server_url(VLACompatLittleEndianStruct):
        """URL of the NTP server to use for time synchronisation"""

        NAME = "NTP_SERVER_URL"
        BASE_ID = 30
        RANGE = 1
        _fields_ = []
        vla_field = ("url", structs.kv_string)
        _pack_ = 1

    class epacket_udp_url(VLACompatLittleEndianStruct):
        """ePacket UDP server hostname"""

        NAME = "EPACKET_UDP_URL"
        BASE_ID = 31
        RANGE = 1
        _fields_ = []
        vla_field = ("server", structs.kv_string)
        _pack_ = 1

    class epacket_udp_port(VLACompatLittleEndianStruct):
        """ePacket UDP server port"""

        NAME = "EPACKET_UDP_PORT"
        BASE_ID = 32
        RANGE = 1
        _fields_ = [
            ("port", ctypes.c_uint32),
        ]
        _pack_ = 1

    class lte_modem_model(VLACompatLittleEndianStruct):
        """Modem model as returned by AT+CGMM"""

        NAME = "LTE_MODEM_MODEL"
        BASE_ID = 40
        RANGE = 1
        _fields_ = []
        vla_field = ("model", structs.kv_string)
        _pack_ = 1

    class lte_modem_firmware_revision(VLACompatLittleEndianStruct):
        """Modem firmware revision as returned by AT+CGMR"""

        NAME = "LTE_MODEM_FIRMWARE_REVISION"
        BASE_ID = 41
        RANGE = 1
        _fields_ = []
        vla_field = ("revision", structs.kv_string)
        _pack_ = 1

    class lte_modem_esn(VLACompatLittleEndianStruct):
        """'Electronic Serial Number' as returned by AT+CGSN=0"""

        NAME = "LTE_MODEM_ESN"
        BASE_ID = 42
        RANGE = 1
        _fields_ = []
        vla_field = ("esn", structs.kv_string)
        _pack_ = 1

    class lte_modem_imei(VLACompatLittleEndianStruct):
        """'International Modem Equipment Identifier' as returned by AT+CGSN=1"""

        NAME = "LTE_MODEM_IMEI"
        BASE_ID = 43
        RANGE = 1
        _fields_ = [
            ("imei", ctypes.c_uint64),
        ]
        _pack_ = 1

    class lte_sim_uicc(VLACompatLittleEndianStruct):
        """SIM Universal Identifier (https://www.itu.int/en/ITU-T/inr/forms/Pages/iin.aspx)"""

        NAME = "LTE_SIM_UICC"
        BASE_ID = 44
        RANGE = 1
        _fields_ = []
        vla_field = ("uicc", structs.kv_string)
        _pack_ = 1

    class lte_pdp_config(VLACompatLittleEndianStruct):
        """Packet Data Protocol (PDP) default context configuration"""

        NAME = "LTE_PDP_CONFIG"
        BASE_ID = 45
        RANGE = 1
        _fields_ = [
            ("family", ctypes.c_uint8),
        ]
        vla_field = ("apn", structs.kv_string)
        _pack_ = 1

    class lte_networking_modes(VLACompatLittleEndianStruct):
        """Enabled LTE networking modes and preferences"""

        NAME = "LTE_NETWORKING_MODES"
        BASE_ID = 46
        RANGE = 1
        _fields_ = [
            ("modes", ctypes.c_uint8),
            ("prefer", ctypes.c_uint8),
        ]
        _pack_ = 1

    class lte_sim_imsi(VLACompatLittleEndianStruct):
        """'International Modem Subscriber Identity' as returned by AT+CIMI"""

        NAME = "LTE_SIM_IMSI"
        BASE_ID = 47
        RANGE = 1
        _fields_ = [
            ("imsi", ctypes.c_uint64),
        ]
        _pack_ = 1

    class bluetooth_peer(VLACompatLittleEndianStruct):
        """Bluetooth peer device"""

        NAME = "BLUETOOTH_PEER"
        BASE_ID = 50
        RANGE = 1
        _fields_ = [
            ("address", structs.bt_addr_le),
        ]
        _pack_ = 1

    class lora_config(VLACompatLittleEndianStruct):
        """LoRa modem configuration"""

        NAME = "LORA_CONFIG"
        BASE_ID = 51
        RANGE = 1
        _fields_ = [
            ("frequency", ctypes.c_uint32),
            ("bandwidth", ctypes.c_uint8),
            ("spreading_factor", ctypes.c_uint8),
            ("coding_rate", ctypes.c_uint8),
            ("preamble_len", ctypes.c_uint16),
            ("tx_power", ctypes.c_int8),
            ("sync_word", ctypes.c_uint8),
        ]
        _pack_ = 1

    class bluetooth_throughput_limit(VLACompatLittleEndianStruct):
        """Request connected Bluetooth peers to limit throughput"""

        NAME = "BLUETOOTH_THROUGHPUT_LIMIT"
        BASE_ID = 52
        RANGE = 1
        _fields_ = [
            ("limit_kbps", ctypes.c_uint16),
        ]
        _pack_ = 1

    class led_disable_daily_time_range(VLACompatLittleEndianStruct):
        """Disable LEDs between two UTC times daily"""

        NAME = "LED_DISABLE_DAILY_TIME_RANGE"
        BASE_ID = 53
        RANGE = 1
        _fields_ = [
            ("disable_start", structs.kv_utc_hms),
            ("disable_end", structs.kv_utc_hms),
        ]
        _pack_ = 1

    class memfault_disable(VLACompatLittleEndianStruct):
        """Disable Memfault reporting at runtime"""

        NAME = "MEMFAULT_DISABLE"
        BASE_ID = 54
        RANGE = 1
        _fields_ = [
            ("disable", ctypes.c_uint8),
        ]
        _pack_ = 1

    class gravity_reference(VLACompatLittleEndianStruct):
        """Reference gravity vector for tilt calculations"""

        NAME = "GRAVITY_REFERENCE"
        BASE_ID = 60
        RANGE = 1
        _fields_ = [
            ("x", ctypes.c_int16),
            ("y", ctypes.c_int16),
            ("z", ctypes.c_int16),
        ]
        _pack_ = 1

    class geofence(VLACompatLittleEndianStruct):
        """Array of points defining a closed polygon"""

        NAME = "GEOFENCE"
        BASE_ID = 100
        RANGE = 16
        _fields_ = [
            ("points_num", ctypes.c_uint8),
        ]
        vla_field = ("points", 0 * structs.gcs_location)
        _pack_ = 1

    class alg_stationary_windowed_args(VLACompatLittleEndianStruct):
        """Configuration for the 'Stationary Windowed' algorithm"""

        NAME = "ALG_STATIONARY_WINDOWED_ARGS"
        BASE_ID = 200
        RANGE = 1
        _fields_ = [
            ("logging", structs.kv_algorithm_logging),
            ("args", structs.kv_algorithm_stationary_windowed_args),
        ]
        _pack_ = 1

    class alg_tilt_args(VLACompatLittleEndianStruct):
        """Configuration for the 'Tilt' algorithm"""

        NAME = "ALG_TILT_ARGS"
        BASE_ID = 201
        RANGE = 1
        _fields_ = [
            ("logging", structs.kv_algorithm_logging),
            ("args", structs.kv_algorithm_tilt_args),
        ]
        _pack_ = 1

    class task_schedules_default_id(VLACompatLittleEndianStruct):
        """Unique identifier for default schedule set"""

        NAME = "TASK_SCHEDULES_DEFAULT_ID"
        BASE_ID = 1000
        RANGE = 1
        _fields_ = [
            ("set_id", ctypes.c_uint32),
        ]
        _pack_ = 1

    class task_schedules(VLACompatLittleEndianStruct):
        """Task runner task schedule definition (@ref task_schedule)"""

        NAME = "TASK_SCHEDULES"
        BASE_ID = 1001
        RANGE = 32
        _fields_ = [
            ("task_id", ctypes.c_uint8),
            ("validity", ctypes.c_uint8),
            ("periodicity_type", ctypes.c_uint8),
            ("boot_lockout_minutes", ctypes.c_uint8),
            ("timeout_s", ctypes.c_uint32),
            ("battery_start", structs.kv_range_u8),
            ("battery_terminate", structs.kv_range_u8),
        ]
        vla_field = ("_remainder", 0 * ctypes.c_uint8)
        _pack_ = 1

    class secure_storage_reserved(VLACompatLittleEndianStruct):
        """Keys reserved for secure storage (do not enable)"""

        NAME = "SECURE_STORAGE_RESERVED"
        BASE_ID = 30000
        RANGE = 10
        _fields_ = [
            ("data_num", ctypes.c_uint8),
        ]
        vla_field = ("data", 0 * ctypes.c_uint8)
        _pack_ = 1

    ID_MAPPING = {
        0: reboots,
        1: bluetooth_addr,
        2: exfat_disk_info,
        3: bluetooth_ctlr_version,
        4: device_name,
        5: infuse_application_id,
        6: application_active,
        7: board_target,
        10: fixed_location,
        20: wifi_ssid,
        21: wifi_psk,
        22: wifi_channels,
        30: ntp_server_url,
        31: epacket_udp_url,
        32: epacket_udp_port,
        40: lte_modem_model,
        41: lte_modem_firmware_revision,
        42: lte_modem_esn,
        43: lte_modem_imei,
        44: lte_sim_uicc,
        45: lte_pdp_config,
        46: lte_networking_modes,
        47: lte_sim_imsi,
        50: bluetooth_peer,
        51: lora_config,
        52: bluetooth_throughput_limit,
        53: led_disable_daily_time_range,
        54: memfault_disable,
        60: gravity_reference,
        100: geofence,
        101: geofence,
        102: geofence,
        103: geofence,
        104: geofence,
        105: geofence,
        106: geofence,
        107: geofence,
        108: geofence,
        109: geofence,
        110: geofence,
        111: geofence,
        112: geofence,
        113: geofence,
        114: geofence,
        115: geofence,
        200: alg_stationary_windowed_args,
        201: alg_tilt_args,
        1000: task_schedules_default_id,
        1001: task_schedules,
        1002: task_schedules,
        1003: task_schedules,
        1004: task_schedules,
        1005: task_schedules,
        1006: task_schedules,
        1007: task_schedules,
        1008: task_schedules,
        1009: task_schedules,
        1010: task_schedules,
        1011: task_schedules,
        1012: task_schedules,
        1013: task_schedules,
        1014: task_schedules,
        1015: task_schedules,
        1016: task_schedules,
        1017: task_schedules,
        1018: task_schedules,
        1019: task_schedules,
        1020: task_schedules,
        1021: task_schedules,
        1022: task_schedules,
        1023: task_schedules,
        1024: task_schedules,
        1025: task_schedules,
        1026: task_schedules,
        1027: task_schedules,
        1028: task_schedules,
        1029: task_schedules,
        1030: task_schedules,
        1031: task_schedules,
        1032: task_schedules,
        30000: secure_storage_reserved,
        30001: secure_storage_reserved,
        30002: secure_storage_reserved,
        30003: secure_storage_reserved,
        30004: secure_storage_reserved,
        30005: secure_storage_reserved,
        30006: secure_storage_reserved,
        30007: secure_storage_reserved,
        30008: secure_storage_reserved,
        30009: secure_storage_reserved,
    }


__all__ = [
    "structs",
    "slots",
]



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.rpc_params import RPCParams
    from ..models.rpc_req_data_header import RPCReqDataHeader


T = TypeVar("T", bound="NewRPCReq")


@_attrs_define
class NewRPCReq:
    """
    Attributes:
        command_id (Union[Unset, int]): ID of RPC command (must provide either commandId or commandName) Example: 3.
        command_name (Union[Unset, str]): Name of RPC command (must provide either commandId or commandName) Example:
            time_get.
        params (Union[Unset, RPCParams]): RPC request or response params (must be a JSON object with string or embedded
            json values - numbers sent as decimal strings) Example: {'primitive_vaue': '1000', 'struct_value': {'field':
            'value'}}.
        params_encoded (Union[Unset, str]): Base64 encoded params (if provided, will be used instead of params)
        data_header (Union[Unset, RPCReqDataHeader]):
    """

    command_id: Unset | int = UNSET
    command_name: Unset | str = UNSET
    params: Union[Unset, "RPCParams"] = UNSET
    params_encoded: Unset | str = UNSET
    data_header: Union[Unset, "RPCReqDataHeader"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        command_id = self.command_id

        command_name = self.command_name

        params: Unset | dict[str, Any] = UNSET
        if not isinstance(self.params, Unset):
            params = self.params.to_dict()

        params_encoded = self.params_encoded

        data_header: Unset | dict[str, Any] = UNSET
        if not isinstance(self.data_header, Unset):
            data_header = self.data_header.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if command_id is not UNSET:
            field_dict["commandId"] = command_id
        if command_name is not UNSET:
            field_dict["commandName"] = command_name
        if params is not UNSET:
            field_dict["params"] = params
        if params_encoded is not UNSET:
            field_dict["paramsEncoded"] = params_encoded
        if data_header is not UNSET:
            field_dict["dataHeader"] = data_header

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.rpc_params import RPCParams
        from ..models.rpc_req_data_header import RPCReqDataHeader

        d = dict(src_dict)
        command_id = d.pop("commandId", UNSET)

        command_name = d.pop("commandName", UNSET)

        _params = d.pop("params", UNSET)
        params: Unset | RPCParams
        if isinstance(_params, Unset):
            params = UNSET
        else:
            params = RPCParams.from_dict(_params)

        params_encoded = d.pop("paramsEncoded", UNSET)

        _data_header = d.pop("dataHeader", UNSET)
        data_header: Unset | RPCReqDataHeader
        if isinstance(_data_header, Unset):
            data_header = UNSET
        else:
            data_header = RPCReqDataHeader.from_dict(_data_header)

        new_rpc_req = cls(
            command_id=command_id,
            command_name=command_name,
            params=params,
            params_encoded=params_encoded,
            data_header=data_header,
        )

        new_rpc_req.additional_properties = d
        return new_rpc_req

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3

import enum
import uuid


class InfuseType(enum.IntEnum):
    """Infuse Data Types"""

    ECHO_REQ = 0
    ECHO_RSP = 1
    TDF = 2
    RPC_CMD = 3
    RPC_DATA = 4
    RPC_DATA_ACK = 5
    RPC_RSP = 6
    RECEIVED_EPACKET = 7
    ACK = 8
    EPACKET_FORWARD = 9
    SERIAL_LOG = 10
    MEMFAULT_CHUNK = 30

    KEY_IDS = 127


class InfuseID(enum.IntEnum):
    """Hardcoded Infuse IDs"""

    GATEWAY = -1


class InfuseBluetoothUUID:
    SERVICE_UUID = uuid.UUID("0000fc74-0000-1000-8000-00805f9b34fb")
    COMMAND_CHAR = uuid.UUID("dc0b71b7-fc74-fc74-aa01-8aba434a893d")
    DATA_CHAR = uuid.UUID("dc0b71b7-fc74-fc74-aa02-8aba434a893d")
    LOGGING_CHAR = uuid.UUID("dc0b71b7-fc74-fc74-aa03-8aba434a893d")



from enum import Enum


class RouteType(str, Enum):
    BT_ADV = "bt_adv"
    BT_CENTRAL = "bt_central"
    BT_PERIPHERAL = "bt_peripheral"
    HCI = "hci"
    SERIAL = "serial"
    UDP = "udp"
    UNKNOWN = "unknown"

    def __str__(self) -> str:
        return str(self.value)



#!/usr/bin/env python3

import ctypes
import os

import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.rpc_wrappers.kv_write import kv_write
from infuse_iot.util.argparse import BtLeAddress
from infuse_iot.util.ctypes import VLACompatLittleEndianStruct
from infuse_iot.zephyr.errno import errno


class kv_bt_peer(InfuseRpcCommand, defs.kv_write):
    HELP = "Configure the peer Bluetooth address"
    DESCRIPTION = "Configure the peer Bluetooth address"

    class request(ctypes.LittleEndianStructure):
        _fields_ = [
            ("num", ctypes.c_uint8),
        ]
        _pack_ = 1

    class response(VLACompatLittleEndianStruct):
        vla_field = ("rc", 0 * ctypes.c_int16)

    @classmethod
    def add_parser(cls, parser):
        addr_group = parser.add_mutually_exclusive_group(required=True)
        addr_group.add_argument("--public", type=BtLeAddress, help="Public Bluetooth address")
        addr_group.add_argument("--random", type=BtLeAddress, help="Random Bluetooth address")
        addr_group.add_argument("--delete", action="store_true", help="Delete peer device")

    def __init__(self, args):
        if args.public:
            addr = bytes(BtLeAddress.to_ctype(defs.rpc_enum_bt_le_addr_type.PUBLIC, args.public))
        elif args.random:
            addr = bytes(BtLeAddress.to_ctype(defs.rpc_enum_bt_le_addr_type.RANDOM, args.random))
        elif args.delete:
            addr = b""
        else:
            raise NotImplementedError
        self.addr = addr

    def request_struct(self):
        addr_struct = kv_write.kv_store_value_factory(50, self.addr)
        request_bytes = bytes(addr_struct)
        return bytes(self.request(1)) + request_bytes

    def handle_response(self, return_code, response):
        if return_code != 0:
            print(f"Invalid data buffer ({errno.strerror(-return_code)})")
            return

        def print_status(name, rc):
            if rc < 0:
                if self.addr == b"":
                    print(f"{name} failed to delete ({os.strerror(-rc)})")
                else:
                    print(f"{name} failed to write ({os.strerror(-rc)})")
            elif rc == 0:
                if self.addr == b"":
                    print(f"{name} deleted")
                else:
                    print(f"{name} already matched")
            else:
                print(f"{name} updated")

        print_status("Peer address", response.rc[0])



#!/usr/bin/env python3

import ctypes
from collections.abc import Generator

from typing_extensions import Any, Self

UINT8_MAX = 2**8 - 1
UINT16_MAX = 2**16 - 1
UINT32_MAX = 2**32 - 1
UINT64_MAX = 2**64 - 1
INT8_MIN = -(2**7)
INT16_MIN = -(2**15)
INT32_MIN = -(2**31)
UINT64_MIN = -(2**63)
INT8_MAX = 2**7 - 1
INT16_MAX = 2**15 - 1
INT32_MAX = 2**31 - 1
INT64_MAX = 2**63 - 1


def bytes_to_uint8(b: bytes):
    return (len(b) * ctypes.c_uint8)(*b)


class VLACompatLittleEndianStruct(ctypes.LittleEndianStructure):
    """
    Class to simplify working with variable length arrays.
    The standard `ctypes.LittleEndianStructure` field definitions can
    be extended with an optional `vla_field` class parameter, which
    will be populated when constructed with `vla_from_buffer_copy`.
    """

    vla_field: tuple[str, type[Any]] | None = None
    vla_counted_by: str | None = None

    @classmethod
    def vla_from_buffer_copy(cls, source, offset=0) -> Self:
        """
        Equivalent to `from_buffer_copy`, except if the `vla_field`
        class property is not `None`, it will consume the remainder of
        `source` to populate the trailing variable length array.
        """

        base = cls.from_buffer_copy(source, offset)
        vla_val: list | VLACompatLittleEndianStruct
        if cls.vla_field is None:
            return base

        remainder = source[ctypes.sizeof(cls) :]
        vla_field_name, vla_field_type = cls.vla_field  # type: ignore

        if issubclass(vla_field_type, ctypes.Array):
            array_base: ctypes._PyCSimpleType = vla_field_type._type_  # type: ignore
            if hasattr(array_base, "vla_counted_by"):
                # This is an array of VLA arrays where the sub-arrys define their own length
                vla_val = []
                # Consume all remaining buffer bytes
                while len(remainder) > 0:
                    sub_vla_field_name, sub_vla_field_type = array_base.vla_field  # type: ignore
                    sub_array_base: ctypes._CData = sub_vla_field_type._type_  # type: ignore
                    sub_base = array_base.from_buffer_copy(remainder)
                    sub_base_size = ctypes.sizeof(sub_base)
                    sub_count = getattr(sub_base, array_base.vla_counted_by)
                    if sub_count < 0:
                        # Assume that negative length is an error code and use 0
                        vla_val.append(sub_base)
                        sub_vla_size = 0
                    else:
                        sub_vla_type = sub_count * sub_array_base
                        # Don't use ctypes.sizeof on constructed type, it returns the wrong value
                        sub_vla_size = sub_count * ctypes.sizeof(sub_array_base)
                        sub_vla_val = sub_vla_type.from_buffer_copy(remainder[sub_base_size:])
                        setattr(sub_base, sub_vla_field_name, sub_vla_val)
                        vla_val.append(sub_base)
                    remainder = remainder[sub_base_size + sub_vla_size :]
            else:
                # Determine the number of VLA elements on "source"
                vla_byte_len = (len(source) - offset) - ctypes.sizeof(cls)
                vla_element_size = ctypes.sizeof(array_base)
                if vla_byte_len % vla_element_size != 0:
                    raise TypeError(f"Unaligned VLA buffer for {cls} (len {len(source)})")
                vla_num = vla_byte_len // vla_element_size
                vla_type = vla_num * array_base
                vla_val = vla_type.from_buffer_copy(remainder)

        elif issubclass(vla_field_type, VLACompatLittleEndianStruct):
            vla_val = vla_field_type.vla_from_buffer_copy(remainder)
        else:
            raise RuntimeError(f"Unhandled VLA type {vla_field_type}")

        setattr(base, vla_field_name, vla_val)
        return base

    def iter_fields(self, prefix: str = "") -> Generator[tuple[str, Any], None, None]:
        """
        Yield all fields of the class as `("name", value)`. Fields that are themselves
        instances of `VLACompatLittleEndianStruct` will have their fields yielded
        individually as `("name.subfield", value)`.
        """
        for field_name, _field_type in self._fields_:  # type: ignore
            val = getattr(self, field_name)
            if isinstance(val, VLACompatLittleEndianStruct):
                yield from val.iter_fields(f"{field_name}.")
            else:
                yield (f"{prefix}{field_name}", val)
        if vla_field := self.vla_field:
            val = getattr(self, vla_field[0])
            if isinstance(val, VLACompatLittleEndianStruct):
                yield from val.iter_fields(f"{vla_field[0]}.")
            else:
                yield (f"{prefix}{vla_field[0]}", val)



#!/usr/bin/env python3

"""Save received TDFs in CSV files"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2024, Embeint Holdings Pty Ltd"

import os
import time

from infuse_iot.commands import InfuseCommand
from infuse_iot.common import InfuseType
from infuse_iot.socket_comms import (
    ClientNotificationEpacketReceived,
    LocalClient,
    default_multicast_address,
)
from infuse_iot.tdf import TDF
from infuse_iot.time import InfuseTime


def _to_str(unix_time: float) -> str:
    return str(unix_time)


class SubCommand(InfuseCommand):
    NAME = "tdf_csv"
    HELP = "Save received TDFs in CSV files"
    DESCRIPTION = "Save received TDFs in CSV files"

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--unix", action="store_true", help="Save timestamps as unix")

    def __init__(self, args):
        self._client = LocalClient(default_multicast_address(), 1.0)
        self._decoder = TDF()
        self.args = args

    def run(self):
        files = {}

        while True:
            msg = self._client.receive()
            if msg is None:
                continue
            if not isinstance(msg, ClientNotificationEpacketReceived):
                continue
            if msg.epacket.ptype != InfuseType.TDF:
                continue
            source = msg.epacket.route[0]

            for tdf in self._decoder.decode(msg.epacket.payload):
                # Construct reading strings
                lines = []
                reading_time = tdf.time
                for idx, reading in enumerate(tdf.data):
                    if self.args.unix:
                        time_func = _to_str
                    else:
                        time_func = InfuseTime.utc_time_string_log

                    if tdf.base_idx is not None:
                        if reading_time is None or idx > 0:
                            time_str = f"{tdf.base_idx + idx}"
                        else:
                            time_str = time_func(reading_time)
                    elif reading_time is None:
                        # Log with local time
                        time_str = time_func(time.time())
                    else:
                        time_str = time_func(reading_time)
                    line = time_str + "," + ",".join([f.val_fmt() for f in reading.iter_fields()])
                    lines.append(line)
                    if tdf.period is not None:
                        assert reading_time is not None
                        reading_time += tdf.period

                # Handle file creation/opening
                first = tdf.data[0]
                filename = f"{source.infuse_id:016x}_{first.NAME}.csv"
                if filename not in files:
                    if os.path.exists(filename):
                        print(f"Appending to existing {filename}")
                        files[filename] = open(filename, "a", encoding="utf-8")  # noqa: SIM115
                    else:
                        print(f"Opening new {filename}")
                        files[filename] = open(filename, "w", encoding="utf-8")  # noqa: SIM115
                        headings = "time," + ",".join([f.name for f in first.iter_fields()])
                        files[filename].write(headings + os.linesep)

                # Write line to file then flush
                for line in lines:
                    files[filename].write(line + os.linesep)
                files[filename].flush()



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK

"""Infuse-IoT SDK meta-tool (infuse) main module"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2024, Embeint Holdings Pty Ltd"

import argparse
import importlib.util
import pathlib
import pkgutil
import sys
import types

import argcomplete

import infuse_iot.tools
from infuse_iot.commands import InfuseCommand
from infuse_iot.credentials import get_custom_tool_path
from infuse_iot.version import __version__


class InfuseApp:
    """The infuse 'application' object"""

    def __init__(self):
        self.args = None
        self.parser = argparse.ArgumentParser("infuse")
        self.parser.add_argument("--version", action="version", version=f"{__version__}")
        self._tools = {}
        # Load tools
        self._load_tools(self.parser)
        # Handle CLI tab completion
        argcomplete.autocomplete(self.parser)

    def run(self, argv):
        """Run the chosen subtool handler"""
        self.args = self.parser.parse_args(argv)

        tool = self.args.tool_class(self.args)
        tool.run()

    def _load_from_module(self, parent_parser: argparse._SubParsersAction, module: types.ModuleType):
        tool_cls: InfuseCommand = module.SubCommand
        parser = parent_parser.add_parser(
            tool_cls.NAME,
            help=tool_cls.HELP,
            description=tool_cls.DESCRIPTION,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        parser.set_defaults(tool_class=tool_cls)
        tool_cls.add_parser(parser)

    def _load_tools(self, parser: argparse.ArgumentParser):
        tools_parser = parser.add_subparsers(title="commands", metavar="<command>", required=True)

        # Iterate over local tools
        for _, name, _ in pkgutil.walk_packages(infuse_iot.tools.__path__):
            full_name = f"{infuse_iot.tools.__name__}.{name}"
            module = importlib.import_module(full_name)
            self._load_from_module(tools_parser, module)

        # Load custom tools, if configured
        if extension_tools := get_custom_tool_path():
            extension_path = pathlib.Path(extension_tools)
            for _, name, _ in pkgutil.walk_packages([extension_tools]):
                full_name = f"{infuse_iot.tools.__name__}.{name}"
                full_path = str(extension_path / f"{name}.py")
                spec = importlib.util.spec_from_file_location(full_name, full_path)
                if spec is None or spec.loader is None:
                    continue
                module = importlib.util.module_from_spec(spec)
                try:
                    spec.loader.exec_module(module)
                except Exception as e:
                    print(f"Failed to import '{name}': {str(e)}")
                    continue
                if hasattr(module, "SubCommand"):
                    self._load_from_module(tools_parser, module)


def main(argv=None):
    """Create the InfuseApp instance and let it run"""
    app = InfuseApp()
    try:
        app.run(argv or sys.argv[1:])
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



#!/usr/bin/env

import ctypes
from abc import ABCMeta, abstractmethod


class ProvisioningInterface(metaclass=ABCMeta):
    "Generic SoC provisioning interface"

    class DefaultProvisioningStruct(ctypes.LittleEndianStructure):
        _fields_ = [
            ("device_id", ctypes.c_uint64),
        ]
        _pack_ = 1

    @property
    @abstractmethod
    def soc_name(self) -> str:
        "Infuse-IoT SoC name"

    @property
    @abstractmethod
    def unique_device_id_len(self) -> int:
        "Length of the unique device ID in bytes"

    @abstractmethod
    def unique_device_id(self) -> int:
        """Read unique device ID"""

    @abstractmethod
    def read_provisioned_data(self, num: int) -> bytes:
        """Read currently provisioned data from device"""

    @abstractmethod
    def write_provisioning_data(self, data: bytes):
        """Write provisioning data to device"""

    @abstractmethod
    def close(self) -> None:
        "Close any interfaces opened"



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.definitions_field_display_fmt import DefinitionsFieldDisplayFmt
from ..types import UNSET, Unset

T = TypeVar("T", bound="DefinitionsFieldDisplay")


@_attrs_define
class DefinitionsFieldDisplay:
    """Display settings for a field

    Attributes:
        fmt (Union[Unset, DefinitionsFieldDisplayFmt]): Format string for field
        digits (Union[Unset, int]):
        postfix (Union[Unset, str]):
    """

    fmt: Unset | DefinitionsFieldDisplayFmt = UNSET
    digits: Unset | int = UNSET
    postfix: Unset | str = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        fmt: Unset | str = UNSET
        if not isinstance(self.fmt, Unset):
            fmt = self.fmt.value

        digits = self.digits

        postfix = self.postfix

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if fmt is not UNSET:
            field_dict["fmt"] = fmt
        if digits is not UNSET:
            field_dict["digits"] = digits
        if postfix is not UNSET:
            field_dict["postfix"] = postfix

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        _fmt = d.pop("fmt", UNSET)
        fmt: Unset | DefinitionsFieldDisplayFmt
        if isinstance(_fmt, Unset):
            fmt = UNSET
        else:
            fmt = DefinitionsFieldDisplayFmt(_fmt)

        digits = d.pop("digits", UNSET)

        postfix = d.pop("postfix", UNSET)

        definitions_field_display = cls(
            fmt=fmt,
            digits=digits,
            postfix=postfix,
        )

        definitions_field_display.additional_properties = d
        return definitions_field_display

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



#!/usr/bin/env python3


import infuse_iot.definitions.rpc as defs
from infuse_iot.commands import InfuseRpcCommand
from infuse_iot.definitions.rpc import (
    rpc_enum_bt_le_addr_type,
    rpc_enum_infuse_bt_characteristic,
    rpc_struct_bt_addr_le,
)
from infuse_iot.util.argparse import BtLeAddress
from infuse_iot.util.ctypes import bytes_to_uint8
from infuse_iot.zephyr.errno import errno
from infuse_iot.zephyr.hci import error


class bt_connect_infuse(InfuseRpcCommand, defs.bt_connect_infuse):
    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--timeout", type=int, default=5000, help="Connection timeout (ms)")
        parser.add_argument("--inactivity", type=int, default=0, help="Data inactivity timeout (ms)")
        parser.add_argument("--data", action="store_true", help="Subscribe to data characteristic")
        parser.add_argument(
            "--logging",
            action="store_true",
            help="Subscribe to serial logging characteristic",
        )
        addr_group = parser.add_mutually_exclusive_group(required=True)
        addr_group.add_argument("--public", type=BtLeAddress, help="Public Bluetooth address")
        addr_group.add_argument("--random", type=BtLeAddress, help="Random Bluetooth address")

    def __init__(self, args):
        self.args = args

    def request_struct(self) -> defs.bt_connect_infuse.request:
        if self.args.public:
            peer = rpc_struct_bt_addr_le(
                rpc_enum_bt_le_addr_type.PUBLIC,
                bytes_to_uint8(self.args.public.to_bytes(6, "little")),
            )
        else:
            peer = rpc_struct_bt_addr_le(
                rpc_enum_bt_le_addr_type.RANDOM,
                bytes_to_uint8(self.args.random.to_bytes(6, "little")),
            )

        # Requested characteristic subscriptions
        sub: int = rpc_enum_infuse_bt_characteristic.COMMAND
        if self.args.data:
            sub |= rpc_enum_infuse_bt_characteristic.DATA
        if self.args.logging:
            sub |= rpc_enum_infuse_bt_characteristic.LOGGING

        return self.request(
            peer,
            self.args.timeout,
            sub,
            self.args.inactivity,
        )

    def handle_response(self, return_code, response):
        if return_code < 0:
            print(f"Failed to connect ({errno.strerror(-return_code)})")
            return
        elif return_code > 0:
            print(f"Failed to connect ({error.strerror(return_code)})")

        print("Connected")
        print(f"\tDevice Public Key: {bytes(response.device_public_key).hex()}")
        print(f"\t Cloud Public Key: {bytes(response.cloud_public_key).hex()}")
        print(f"\t          Network: 0x{response.network_id:06x}")



#!/usr/bin/env python3

"""Run remote procedure calls on devices"""

__author__ = "Jordan Yates"
__copyright__ = "Copyright 2024, Embeint Holdings Pty Ltd"

import argparse
import importlib
import pkgutil
import random

import infuse_iot.rpc_wrappers as wrappers
from infuse_iot.commands import InfuseCommand, InfuseRpcCommand
from infuse_iot.common import InfuseID, InfuseType
from infuse_iot.rpc_client import RpcClient
from infuse_iot.socket_comms import (
    ClientNotification,
    ClientNotificationEpacketReceived,
    GatewayRequestConnectionRequest,
    LocalClient,
    default_multicast_address,
)


class SubCommand(InfuseCommand):
    NAME = "rpc"
    HELP = "Run remote procedure calls on devices"
    DESCRIPTION = "Run remote procedure calls on devices"

    @classmethod
    def add_parser(cls, parser):
        addr_group = parser.add_mutually_exclusive_group(required=True)
        addr_group.add_argument("--gateway", action="store_true", help="Run command on local gateway")
        addr_group.add_argument("--id", type=lambda x: int(x, 0), help="Infuse ID to run command on")
        parser.add_argument("--conn-log", action="store_true", help="Request logs from remote device")
        parser.add_argument(
            "--conn-timeout", type=int, default=10000, help="Timeout to wait for a connection to the device (ms)"
        )
        command_list_parser = parser.add_subparsers(title="commands", metavar="<command>", required=True)

        for _, name, _ in pkgutil.walk_packages(wrappers.__path__):
            full_name = f"{wrappers.__name__}.{name}"
            module = importlib.import_module(full_name)

            # Add RPC wrapper to parser
            cmd_cls = getattr(module, name)
            cmd_parser = command_list_parser.add_parser(
                name,
                help=cmd_cls.HELP,
                description=cmd_cls.DESCRIPTION,
                formatter_class=argparse.RawDescriptionHelpFormatter,
            )
            cmd_parser.set_defaults(rpc_class=cmd_cls)
            cmd_cls.add_parser(cmd_parser)

    def __init__(self, args: argparse.Namespace):
        self._args = args
        self._client = LocalClient(default_multicast_address(), 1.0)
        self._command: InfuseRpcCommand = args.rpc_class(args)
        self._request_id = random.randint(0, 2**32 - 1)
        self._max_payload = 0
        if args.gateway:
            self._id = InfuseID.GATEWAY
        else:
            self._id = args.id

    def rx_handler(self, pkt: ClientNotification):
        if (
            self._args.conn_log
            and isinstance(pkt, ClientNotificationEpacketReceived)
            and pkt.epacket.ptype == InfuseType.SERIAL_LOG
        ):
            print(pkt.epacket.payload.decode("utf-8"), end="")

    def run(self):
        try:
            types = GatewayRequestConnectionRequest.DataType.COMMAND
            if self._args.conn_log:
                types |= GatewayRequestConnectionRequest.DataType.LOGGING
            with self._client.connection(self._id, types, self._args.conn_timeout) as mtu:
                self._max_payload = mtu
                rpc_client = RpcClient(self._client, mtu, self._id, self.rx_handler)
                params = bytes(self._command.request_struct())

                if hasattr(self._command.response, "vla_from_buffer_copy"):  # type: ignore
                    decode_fn = self._command.response.vla_from_buffer_copy  # type: ignore
                else:
                    decode_fn = self._command.response.from_buffer_copy  # type: ignore

                if self._command.RPC_DATA_SEND:
                    hdr, rsp = rpc_client.run_data_send_cmd(
                        self._command.COMMAND_ID,  # type: ignore
                        self._command.auth_level(),
                        params,
                        self._command.data_payload(),
                        self._command.data_progress_cb,
                        decode_fn,
                    )
                elif self._command.RPC_DATA_SEND_CHUNKED:
                    hdr, rsp = rpc_client.run_data_send_cmd_chunked(
                        self._command.COMMAND_ID,  # type: ignore
                        self._command.auth_level(),
                        params,
                        self._command.data_payload_chunked(),
                        self._command.data_progress_cb,
                        decode_fn,
                    )
                elif self._command.RPC_DATA_RECEIVE:
                    hdr, rsp = rpc_client.run_data_recv_cmd(
                        self._command.COMMAND_ID,  # type: ignore
                        self._command.auth_level(),
                        params,
                        self._command.data_payload_recv_len(),
                        self._command.data_recv_cb,
                        decode_fn,
                    )
                else:
                    hdr, rsp = rpc_client.run_standard_cmd(
                        self._command.COMMAND_ID,  # type: ignore
                        self._command.auth_level(),
                        params,
                        decode_fn,
                    )
                # Handle response
                self._command.handle_response(hdr.return_code, rsp)

                if self._args.conn_log:
                    while True:
                        if notification := self._client.receive():
                            self.rx_handler(notification)
        except ConnectionRefusedError:
            print(f"Unable to connect to {self._id:016x}")



#!/usr/bin/env python3

import functools
import platform


@functools.cache
def is_wsl(v: str = platform.uname().release) -> bool:
    """
    Are we running under WSL?
    """
    return v.endswith("-Microsoft") or v.endswith("microsoft-standard-WSL2")



#!/usr/bin/env python3

import enum


class IPProtocol(enum.IntEnum):
    IPPROTO_IP = 0
    IPPROTO_ICMP = 1
    IPPROTO_IGMP = 2
    IPPROTO_ETH_P_ALL = 3
    IPPROTO_IPIP = 4
    IPPROTO_TCP = 6
    IPPROTO_UDP = 17
    IPPROTO_IPV6 = 41
    IPPROTO_ICMPV6 = 58
    IPPROTO_RAW = 255


class ProtocolFamily(enum.IntEnum):
    PF_UNSPEC = 0
    PF_INET = 1
    PF_INET6 = 2
    PF_PACKET = 3
    PF_CAN = 4
    PF_NET_MGMT = 5
    PF_LOCAL = 6
    PF_UNIX = PF_LOCAL


class AddressFamily(enum.IntEnum):
    AF_UNSPEC = ProtocolFamily.PF_UNSPEC
    AF_INET = ProtocolFamily.PF_INET
    AF_INET6 = ProtocolFamily.PF_INET6
    AF_PACKET = ProtocolFamily.PF_PACKET
    AF_CAN = ProtocolFamily.PF_CAN
    AF_NET_MGMT = ProtocolFamily.PF_NET_MGMT
    AF_LOCAL = ProtocolFamily.PF_LOCAL
    AF_UNIX = ProtocolFamily.PF_UNIX


class SockType(enum.IntEnum):
    SOCK_STREAM = 1
    SOCK_DGRAM = 2
    SOCK_RAW = 3



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.algorithm import Algorithm
    from ..models.application_version import ApplicationVersion


T = TypeVar("T", bound="NewDeviceState")


@_attrs_define
class NewDeviceState:
    """
    Attributes:
        application_id (Union[Unset, int]): Last announced application ID
        application_version (Union[Unset, ApplicationVersion]): Application version
        algorithms (Union[Unset, list['Algorithm']]): Last announced algorithms
    """

    application_id: Unset | int = UNSET
    application_version: Union[Unset, "ApplicationVersion"] = UNSET
    algorithms: Unset | list["Algorithm"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        application_id = self.application_id

        application_version: Unset | dict[str, Any] = UNSET
        if not isinstance(self.application_version, Unset):
            application_version = self.application_version.to_dict()

        algorithms: Unset | list[dict[str, Any]] = UNSET
        if not isinstance(self.algorithms, Unset):
            algorithms = []
            for algorithms_item_data in self.algorithms:
                algorithms_item = algorithms_item_data.to_dict()
                algorithms.append(algorithms_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if application_id is not UNSET:
            field_dict["applicationId"] = application_id
        if application_version is not UNSET:
            field_dict["applicationVersion"] = application_version
        if algorithms is not UNSET:
            field_dict["algorithms"] = algorithms

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.algorithm import Algorithm
        from ..models.application_version import ApplicationVersion

        d = dict(src_dict)
        application_id = d.pop("applicationId", UNSET)

        _application_version = d.pop("applicationVersion", UNSET)
        application_version: Unset | ApplicationVersion
        if isinstance(_application_version, Unset):
            application_version = UNSET
        else:
            application_version = ApplicationVersion.from_dict(_application_version)

        algorithms = []
        _algorithms = d.pop("algorithms", UNSET)
        for algorithms_item_data in _algorithms or []:
            algorithms_item = Algorithm.from_dict(algorithms_item_data)

            algorithms.append(algorithms_item)

        new_device_state = cls(
            application_id=application_id,
            application_version=application_version,
            algorithms=algorithms,
        )

        new_device_state.additional_properties = d
        return new_device_state

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
